/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster;

import static org.eiichiro.jazzmaster.Version.*;
import static org.eiichiro.jazzmaster.deployment.Stage.*;

import org.eiichiro.jazzmaster.deployment.Deployer;
import org.eiichiro.jazzmaster.deployment.Stage;
import org.eiichiro.jazzmaster.frontend.Frontend;
import org.eiichiro.jazzmaster.frontend.Injection;
import org.eiichiro.jazzmaster.kernel.Kernel;
import org.eiichiro.jazzmaster.management.Configuration;
import org.eiichiro.jazzmaster.management.Manager;
import org.eiichiro.reverb.time.Stopwatch;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code Application} represents the application built on the Jazzmaster 
 * framework.
 * {@code Application} is the entry point to 
 * <a href="http://code.google.com/p/jazzmaster/">Jazzmaster</a>. Components in 
 * Jazzmaster application can be used through this class. However, you have to 
 * bootstrap application invoking {@code #bootstrap()} method before invoking 
 * other method. If you use {@code Application} class when a bootstrap has not 
 * completed or a shutdown has completed or these processes is under processing, 
 * {@code IllegalStateException} will be thrown.
 *
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class Application {

	private static boolean web = false;
	
	private static Logger logger = LoggerFactory.getLogger(Application.class);
	
	private static Bootstrap bootstrap;
	
	private static Shutdown shutdown;
	
	static {
		logger.info("Jazzmaster " + MAJOR + "." + MINER + "." + BUILD);
		logger.info("Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.");
	}

	static void setWeb(boolean web) {
		Application.web = web;
	}

	/**
	 * Returns <code>true</code> if this application is running on web 
	 * application environment.
	 * 
	 * @return <code>true</code> If this application is running on web 
	 * application environment.
	 */
	public static boolean isWeb() {
		return web;
	}
	
	/** Bootstraps in {@code Stage#PRODUCTION} stage. */
	public static void bootstrap() {
		bootstrap(PRODUCTION);
	}
	
	/**
	 * Bootstraps in specified development {@code Stage}.
	 * This invocation is forwarded to 
	 * {@code Application#bootstrap(Stage, Configuration)} with anonymous 
	 * {@code Configuration} implementation that holds on no value.
	 * 
	 * @param stage The development stage this Jazzmaster runs.
	 */
	public static void bootstrap(Stage stage) {
		bootstrap(stage, new Configuration() {

			public Object get(Object key) {
				return null;
			}

			public void set(Object key, Object value) {}
			
		});
	}
	
	/**
	 * Bootstraps with the specified {@code Configuration}.
	 * This invocation is forwarded to 
	 * {@code Application#bootstrap(Stage, Configuration)} with 
	 * {@code Stage#PRODUCTION} stage.
	 * 
	 * @param configuration The configuration for this Jazzmaster application.
	 */
	public static void bootstrap(Configuration configuration) {
		bootstrap(PRODUCTION, configuration);
	}
	
	/**
	 * Bootstraps with the specified development {@code Stage} and 
	 * {@code Configuration}.
	 * 
	 * @param stage The development stage this Jazzmaster application runs.
	 * @param configuration The configuration for this Jazzmaster application.
	 */
	public static synchronized void bootstrap(Stage stage, Configuration configuration) {
		if (bootstrap != null && shutdown == null) {
			logger.warn("Starting shutdown once before rebootstrap");
			shutdown();
		}
		
		logger.info("Starting bootstrap"
				+ ((Application.isWeb()) ? " in web application environment"
						: ""));
		Stopwatch stopwatch = new Stopwatch();
		stopwatch.start();
		Bootstrap bootstrap = new Bootstrap(stage, configuration);
		bootstrap.bootstrap();
		stopwatch.stop();
		Application.bootstrap = bootstrap;
		shutdown = null;
		logger.info("Bootstrap completed in [" + stopwatch + "]");
	}
	
	/** Shutdowns Jazzmaster application. */
	public static synchronized void shutdown() {
		checkState();
		logger.info("Starting shutdown");
		Stopwatch stopwatch = new Stopwatch();
		stopwatch.start();
		Shutdown shutdown = new Shutdown();
		shutdown.shutdown();
		stopwatch.stop();
		Application.shutdown = shutdown;
		bootstrap = null;
		logger.info("Shutdown completed in [" + stopwatch + "]");
	}
	
	/**
	 * Returns frontend instance constructed from the specified class.
	 * The specified class must be annotated {@link Frontend}.
	 * 
	 * @param clazz The frontend class to construct.
	 * @return The frontend instance constructed with specified class.
	 */
	public static <F> F getFrontend(Class<F> clazz) {
		checkState();
		
		if (clazz == null) {
			throw new IllegalArgumentException("'clazz' must not be [" + clazz
					+ "]");
		}
		
		if (clazz.getAnnotation(Frontend.class) == null) {
			throw new IllegalArgumentException("Frontend class [" 
					+ clazz 
					+ "] must be annotated @org.eiichiro.jazzmaster.frontend.Frontend");
		}
		
		return getFrontend(getKernel().assemble(clazz).getInstance());
	}
	
	/**
	 * Returns frontend instance assembled with service dependencies.
	 * 
	 * @param frontend The frontend instance to assemble.
	 * @return The frontend instance assembled with service dependencies.
	 */
	public static <F> F getFrontend(F frontend) {
		checkState();
		
		if (frontend == null) {
			throw new IllegalArgumentException("'frontend' must not be ["
					+ frontend + "]");
		}
		
		if (frontend.getClass().getAnnotation(Frontend.class) == null) {
			throw new IllegalArgumentException("Frontend instance class [" 
					+ frontend 
					+ "] must be annotated @org.eiichiro.jazzmaster.frontend.Frontend");
		}
		
		Injection injection = new Injection();
		injection.inject(frontend);
		return frontend;
	}
	
	public static Kernel getKernel() {
		checkState();
		return bootstrap.getKernel();
	}
	
	public static Manager getManager() {
		checkState();
		return bootstrap.getManager();
	}
	
	public static Deployer getDeployer() {
		checkState();
		return bootstrap.getDeployer();
	}
	
	private static void checkState() {
		if (bootstrap == null) {
			if (shutdown != null) {
				throw new IllegalStateException(
						"Jazzmaster application has been still shutdown");
			} else {
				throw new IllegalStateException(
						"Jazzmaster application hasn't bootstrapped yet");
			}
		}
	}
	
}
