/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster;

import java.io.IOException;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.eiichiro.jazzmaster.management.Scope;

/**
 * {@code WebFilter} holds the current {@code HttpServletRequest} and 
 * {@code HttpServletResponse} in its {@code ThreadLocal} field for 
 * {@code Scope} context processing.
 * {@code Scope#REQUEST} scoped component is managed in the current 
 * {@code HttpServletRequest} attribute and {@code Scope#SESSION} scoped 
 * component is also managed in the current {@code HttpSession} attribute. 
 * {@code Scope#REQUEST} scope's context implementation gets the current 
 * {@code HttpServletRequest} through this filter and stores components into its 
 * attribute. Just the same, {@code Scope#SESSION} scope's context 
 * implementation gets the current {@code HttpSession} through this filter and 
 * stores components into its attribute. Therefore, it is necessary to describe 
 * this filter setting in {@code web.xml} file to use the {@code Scope} in your 
 * web application environment. {@code web.xml} needs the following setting: 
 * <pre>
 * &lt;web-app version="2.5"
 *     xmlns="http://java.sun.com/xml/ns/javaee"
 *     xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
 *     xsi:schemaLocation="http://java.sun.com/xml/ns/javaee
 *         http://java.sun.com/xml/ns/javaee/web-app_2_5.xsd"&gt;
 *     
 *     &lt;!-- Jazzmaster web application. --&gt;
 *     
 *     &lt;filter&gt;
 *         &lt;filter-name&gt;webFilter&lt;/filter-name&gt;
 *         &lt;filter-class&gt;org.eiichiro.jazzmaster.WebFilter&lt;/filter-class&gt;
 *     &lt;/filter&gt;
 *     
 *     &lt;filter-mapping&gt;
 *         &lt;filter-name&gt;webFilter&lt;/filter-name&gt;
 *         &lt;url-pattern&gt;/yourPattern/*&lt;/url-pattern&gt;
 *     &lt;/filter-mapping&gt;
 *     
 * &lt;/web-app&gt;
 * </pre>
 * 
 * @see Scope
 * @see HttpServletRequest
 * @see HttpServletResponse
 *
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class WebFilter implements Filter {

	private static ThreadLocal<HttpServletRequest> httpServletRequest 
			= new ThreadLocal<HttpServletRequest>();
	
	private static ThreadLocal<HttpServletResponse> httpServletResponse 
			= new ThreadLocal<HttpServletResponse>();
	
	/**
	 * No-op.
	 * 
	 * @see javax.servlet.Filter#destroy()
	 */
	public void destroy() {}

	/**
	 * Filtering of every requests.
	 * 
	 * @see javax.servlet.Filter#doFilter(javax.servlet.ServletRequest,
	 *      javax.servlet.ServletResponse, javax.servlet.FilterChain)
	 */
	public void doFilter(ServletRequest servletRequest,
			ServletResponse servletResponse, FilterChain filterChain)
			throws IOException, ServletException {
		HttpServletRequest httpServletRequest 
				= WebFilter.httpServletRequest.get();
		HttpServletResponse httpServletResponse 
				= WebFilter.httpServletResponse.get();
		
		try {
			WebFilter.httpServletRequest.set(HttpServletRequest.class
					.cast(servletRequest));
			WebFilter.httpServletResponse.set(HttpServletResponse.class
					.cast(servletResponse));
			filterChain.doFilter(servletRequest, servletResponse);
		} finally {
			WebFilter.httpServletRequest.set(httpServletRequest);
			WebFilter.httpServletResponse.set(httpServletResponse);
		}
	}
	
	/**
	 * No-op.
	 * 
	 * @see javax.servlet.Filter#init(javax.servlet.FilterConfig)
	 */
	public void init(FilterConfig arg0) throws ServletException {}
	
	/**
	 * Returns the current {@code HttpServletRequest}.
	 * 
	 * @return The current {@code HttpServletRequest}.
	 */
	public static HttpServletRequest getHttpServletRequest() {
		return httpServletRequest.get();
	}
	
	/**
	 * Returns the current {@code HttpServletResponse}.
	 * 
	 * @return The current {@code HttpServletResponse}.
	 */
	public static HttpServletResponse getHttpServletResponse() {
		return httpServletResponse.get();
	}
	
	/**
	 * Returns the current {@code HttpSession}.
	 * 
	 * @return The current {@code HttpSession}.
	 */
	public static HttpSession getHttpSession() {
		HttpServletRequest httpServletRequest 
				= WebFilter.httpServletRequest.get();
		return (httpServletRequest == null) 
				? null : httpServletRequest.getSession();
	}
	
	public static void setHttpServletRequest(
			HttpServletRequest httpServletRequest) {
		WebFilter.httpServletRequest.set(httpServletRequest);
	}

	public static void setHttpServletResponse(
			HttpServletResponse httpServletResponse) {
		WebFilter.httpServletResponse.set(httpServletResponse);
	}


	
}
