/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;
import javax.servlet.ServletRequest;
import javax.servlet.ServletRequestEvent;
import javax.servlet.ServletRequestListener;
import javax.servlet.http.HttpSession;
import javax.servlet.http.HttpSessionEvent;
import javax.servlet.http.HttpSessionListener;

import org.eiichiro.jazzmaster.deployment.Stage;
import org.eiichiro.jazzmaster.management.Configuration;
import org.eiichiro.jazzmaster.management.Lifecycle;
import org.eiichiro.jazzmaster.management.Scope;

/**
 * {@code WebListener} is a {@code ServletContextListener}, 
 * {@code ServletRequestListener} and {@code HttpSessionListener} implementation 
 * to control the {@code Scope#REQUEST} and {@code Scope#SESSION} context.
 * This class is invoked when the servlet context has been initialized and then 
 * bootstraps Jazzmaster application. By using {@code web.xml} (standard web 
 * deployment descriptor), you can also specify the deployment stage and 
 * configuration in web application environment. To specify the deployment stage 
 * and configuration, you need to describe them in your {@code web.xml} as the 
 * following: 
 * <pre>
 * &lt;web-app version="2.5"
 *     xmlns="http://java.sun.com/xml/ns/javaee"
 *     xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
 *     xsi:schemaLocation="http://java.sun.com/xml/ns/javaee
 *         http://java.sun.com/xml/ns/javaee/web-app_2_5.xsd"&gt;
 *     
 *     &lt;!-- Jazzmaster web application. --&gt;
 *     
 *     &lt;listener&gt;
 *         &lt;!-- This class. --&gt;
 *         &lt;listener-class&gt;org.eiichiro.jazzmaster.WebListener&lt;/listener-class&gt;
 *     &lt;/listener&gt;
 *     
 *     &lt;context-param&gt;
 *         &lt;!-- The fixed key for specifying deployment stage. If this key is not 
 *         described, Jazzmaster application is bootstrapped in PRODUCTION stage. --&gt;
 *         &lt;param-name&gt;org.eiichiro.jazzmaster.deployment.stage&lt;/param-name&gt;
 *         &lt;!-- Fully qualified deployment stage constant declaration path. --&gt;
 *         &lt;param-value&gt;TEST&lt;/param-value&gt;
 *     &lt;/context-param&gt;
 *     &lt;context-param&gt;
 *         &lt;!-- The fixed key for specifying Configuration implementation class. 
 *         If this key is not described, Jazzmaster application is bootstrapped with 
 *         no configuration. --&gt;
 *         &lt;param-name&gt;org.eiichiro.jazzmaster.management.configuration&lt;/param-name&gt;
 *         &lt;!-- Fully qualified your Configuration implementation class name. This 
 *         class must have the default constructor. --&gt;
 *         &lt;param-value&gt;com.your.app.YourConfiguration&lt;/param-value&gt;
 *     &lt;/context-param&gt;
 *     
 * &lt;/web-app&gt;
 * </pre>
 * {@code WebListener} is invoked when the servlet context has been destroyed and 
 * shutdowns Jazzmaster application.
 *
 * @see Scope
 * @see Application
 * @see Stage
 * @see Configuration
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class WebListener implements ServletContextListener,
		ServletRequestListener, HttpSessionListener {

	private static final String STAGE 
			= "org.eiichiro.jazzmaster.deployment.stage";
	
	private static final String CONFIGURATION 
			= "org.eiichiro.jazzmaster.management.configuration";
	
	/**
	 * Shutdowns Jazzmaster application in web environment.
	 * This method invokes {@code Shutdown#shutdown()}.
	 * 
	 * @param servletContextEvent The servlet context for the current web 
	 * application.
	 * @see Shutdown#shutdown()
	 * @see javax.servlet.ServletContextListener#contextDestroyed(javax.servlet.ServletContextEvent)
	 */
	public void contextDestroyed(ServletContextEvent arg0) {
		Application.shutdown();
	}

	/**
	 * Bootstraps Jazzmaster application in web environment.
	 * This method loads and assembles deployment stage and 
	 * {@code Configuration} implementation class from {@code web.xml}, and 
	 * invokes {@code Application#bootstrap()} method with them.
	 * 
	 * @param servletContextEvent The servlet context for the current web 
	 * application.
	 * @throws BootstrapException Thrown if this method failed to load and 
	 * assemble deployment stage and configuration implementation or 
	 * {@code Application#bootstrap()} failed.
	 * @see javax.servlet.ServletContextListener#contextInitialized(javax.servlet.ServletContextEvent)
	 * @see Stage
	 * @see Configuration
	 * @see Application#bootstrap()
	 * @see Application#bootstrap(org.eiichiro.jazzmaster.management.Configuration)
	 * @see Application#bootstrap(org.eiichiro.jazzmaster.deployment.Stage)
	 * @see Application#bootstrap(org.eiichiro.jazzmaster.deployment.Stage, org.eiichiro.jazzmaster.management.Configuration)
	 */
	public void contextInitialized(ServletContextEvent servletContextEvent) {
		// For Application static initializer logging.
		Application.class.getClass();
		
		Stage stage = getStage(servletContextEvent);
		Configuration configuration = getConfiguration(servletContextEvent);
		Application.setWeb(true);
		
		if (stage != null && configuration == null) {
			Application.bootstrap(stage);
		} else if (stage == null && configuration != null) {
			Application.bootstrap(configuration);
		} else if (stage != null && configuration != null) {
			Application.bootstrap(stage, configuration);
		} else {
			Application.bootstrap();
		}
	}
	
	private Stage getStage(ServletContextEvent servletContextEvent) {
		String stage = servletContextEvent.getServletContext()
				.getInitParameter(STAGE);
		
		if (stage == null) {
			return null;
		}
		
		try {
			return Stage.valueOf(stage);
		} catch (Exception e) {
			throw new BootstrapException(
					"Web bootstrap failed: Stage declared at web.xml [" + stage
							+ "] cannot be instantiated", e);
		}
	}
	
	protected Configuration getConfiguration(
			ServletContextEvent servletContextEvent) {
		String configuration = servletContextEvent.getServletContext()
				.getInitParameter(CONFIGURATION);
		
		if (configuration == null) {
			return null;
		}
		
		try {
			return Configuration.class.cast(Class.forName(configuration)
					.newInstance());
		} catch (Exception e) {
			throw new BootstrapException(
					"Web bootstrap failed: Configuration declared at web.xml ["
							+ configuration + "] cannot be instantiated", e);
		}
	}
	
	/**
	 * Stops and destroys request scoped components.
	 * This method is invoked when a servlet request is destroyed. This method 
	 * invokes {@code Lifecycle#STOPPED} and {@code Lifecycle#DESTROYED} 
	 * lifecycle callback method on the request scoped component. Then, removes 
	 * all request scoped components from the servlet request.
	 * 
	 * @param servletRequestEvent The event that represents a 
	 * {@code ServletRequest} is destroyed.
	 * @see javax.servlet.ServletRequestListener#requestDestroyed(javax.servlet.ServletRequestEvent)
	 */
	public void requestDestroyed(ServletRequestEvent servletRequestEvent) {
		ServletRequest servletRequest = servletRequestEvent.getServletRequest();
		callback(servletRequest, Lifecycle.STOPPED);
		callback(servletRequest, Lifecycle.DESTROYED);
		
		synchronized (servletRequest) {
			@SuppressWarnings("unchecked")
			Enumeration<String> attributeNames = servletRequest.getAttributeNames();
			List<String> keys = new ArrayList<String>();
			
			while (attributeNames.hasMoreElements()) {
				String nextElement = attributeNames.nextElement();
				
				if (nextElement.startsWith(Scope.REQUEST_ATTRIBUTE)) {
					keys.add(nextElement);
				}
			}
			
			for (String key : keys) {
				servletRequest.removeAttribute(key);
			}
		}
	}
	
	/**
	 * No-op.
	 * 
	 * @param servletRequestEvent The event that represents a 
	 * {@code ServletRequest} is initialized.
	 * @see javax.servlet.ServletRequestListener#requestInitialized(javax.servlet.ServletRequestEvent)
	 */
	public void requestInitialized(ServletRequestEvent servletRequestEvent) {}

	/**
	 * No-op.
	 * 
	 * @param httpSessionEvent The event that represents a 
	 * {@code ServletRequest} is created.
	 * @see javax.servlet.http.HttpSessionListener#sessionCreated(javax.servlet.http.HttpSessionEvent)
	 */
	public void sessionCreated(HttpSessionEvent httpSessionEvent) {}

	/**
	 * Stops and destroys session scoped components.
	 * This method is invoked when a session is destroyed. This method invokes 
	 * {@code Lifecycle#STOPPED} and {@code Lifecycle#DESTROYED} lifecycle 
	 * callback method on the session scoped component. Then, removes all 
	 * session scoped components from the session.
	 * 
	 * @param httpSessionEvent The event that represents a {@code HttpSession} 
	 * is destroyed.
	 * @see javax.servlet.http.HttpSessionListener#sessionDestroyed(javax.servlet.http.HttpSessionEvent)
	 */
	public void sessionDestroyed(HttpSessionEvent httpSessionEvent) {
		HttpSession httpSession = httpSessionEvent.getSession();
		callback(httpSession, Lifecycle.STOPPED);
		callback(httpSession, Lifecycle.DESTROYED);
		
		synchronized (httpSession) {
			@SuppressWarnings("unchecked")
			Enumeration<String> attributeNames = httpSession.getAttributeNames();
			List<String> keys = new ArrayList<String>();
			
			while (attributeNames.hasMoreElements()) {
				String nextElement = attributeNames.nextElement();
				
				if (nextElement.startsWith(Scope.SESSION_ATTRIBUTE)) {
					keys.add(nextElement);
				}
			}
			
			for (String key : keys) {
				httpSession.removeAttribute(key);
			}
		}
	}
	
	private void callback(ServletRequest servletRequest, Lifecycle lifecycle) {
		@SuppressWarnings("unchecked")
		Enumeration<String> attributeNames = servletRequest.getAttributeNames();
		
		while (attributeNames.hasMoreElements()) {
			String nextElement = attributeNames.nextElement();
			
			if (nextElement.startsWith(Scope.REQUEST_ATTRIBUTE)) {
				Application.getManager().callback(servletRequest.getAttribute(nextElement), lifecycle);
			}
		}
	}
	
	private void callback(HttpSession httpSession, Lifecycle lifecycle) {
		@SuppressWarnings("unchecked")
		Enumeration<String> attributeNames = httpSession.getAttributeNames();
		
		while (attributeNames.hasMoreElements()) {
			String nextElement = attributeNames.nextElement();
			
			if (nextElement.startsWith(Scope.SESSION_ATTRIBUTE)) {
				Application.getManager().callback(httpSession.getAttribute(nextElement), lifecycle);
			}
		}
	}
	
}
