/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.bus;

import static org.eiichiro.jazzmaster.management.Lifecycle.CONFIGURED;
import static org.eiichiro.jazzmaster.management.Lifecycle.DESTROYED;
import static org.eiichiro.jazzmaster.management.Lifecycle.LOADED;
import static org.eiichiro.jazzmaster.management.Lifecycle.STARTED;
import static org.eiichiro.jazzmaster.management.Lifecycle.STOPPED;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;

import org.eiichiro.jazzmaster.deployment.Deployer;
import org.eiichiro.jazzmaster.kernel.Injected;
import org.eiichiro.jazzmaster.kernel.Kernel;
import org.eiichiro.jazzmaster.kernel.Property;
import org.eiichiro.jazzmaster.management.Lifecycle;
import org.eiichiro.jazzmaster.management.Manager;
import org.eiichiro.jazzmaster.management.On;
import org.eiichiro.jazzmaster.management.Scope;
import org.eiichiro.jazzmaster.repository.Key;
import org.eiichiro.jazzmaster.service.Service;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code Bus} is the execution and interaction engine of Jazzmaster services.
 * All Jazzmaster services are interacted with each other and executed on the 
 * bus. And, {@code Bus} is also the agent for frontends. Service invocations by 
 * frontend are all dispatched to bus automatically and monitored these 
 * processing. {@code Bus} is consist of the following components: 
 * <ul>
 * <li>Service proxy (Service invocation handler)</li>
 * Jazzmaster service is injected to other services or to frontends 
 * automatically by Jazzmaster's dependency injection. However, frontend or 
 * services as consumer of the other services may not invoke its requiring 
 * service directly. In Jazzmaster application, these injected services are all 
 * proxy (generated with {@link Proxy}) and service invocation is encoded to 
 * bus's dispatcher invocation.
 * <li>Exposer</li>
 * At deployment, loaded and configured service definition ({@link Service} 
 * extension class instance) is passed to exposer. Exposer is running on the bus, 
 * and holding service definition that is being able to expose to frontends and 
 * other services. When a service is required, bus gets the service instance 
 * from exposer, then generates service proxy with its definition. 
 * <li>Dispatcher</li>
 * Service invocations are processed by dispatcher in the bus. Service (proxy) 
 * is invoked once, the bus receives its service definition, invoked method and 
 * parameters. Then, the bus forwards service invocation to dispatcher. The 
 * dispatcher binds interceptors to invoked service method, and executes service 
 * invocation with service instance and parameters.
 * </ul>
 * 
 * @see Service
 * @see Exposer
 * @see Dispatcher
 * @see Interceptor
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class Bus {

	private Logger logger = LoggerFactory.getLogger(Bus.class);
	
	@Injected private Kernel kernel;
	
	@Injected private Manager manager;
	
	private Exposer	exposer;
	
	private Dispatcher dispatcher;
	
	private Set<Class<?>> detected = new CopyOnWriteArraySet<Class<?>>();
	
	/**
	 * Constructs {@code Exposer} and {@code Dispatcher}s. 
	 * Called by manager on this {@code Bus} instance is configured by deployer.
	 * 
	 * @see Deployer
	 * @see Manager
	 * @see Lifecycle
	 */
	@On(CONFIGURED)
	public void configure() {
		Class<Exposer> exposer = Exposer.class;
		manager.setCallback(exposer);
		manager.callback(exposer, LOADED);
		logger.debug("Exposer [" + exposer.getName() + "] loaded");
		this.exposer = kernel.assemble(exposer).getInstance();
		manager.setCallback(this.exposer);
		manager.callback(this.exposer, CONFIGURED);
		
		Class<Dispatcher> dispatcher = Dispatcher.class;
		manager.setCallback(dispatcher);
		manager.callback(dispatcher, LOADED);
		logger.debug("Dispatcher [" + dispatcher.getName() + "] loaded");
		this.dispatcher = kernel.assemble(dispatcher).getInstance();
		manager.setCallback(this.dispatcher);
		manager.callback(this.dispatcher, CONFIGURED);
	}
	
	@On(STARTED)
	public void start() {
		callback(STARTED);
	}
	
	@On(STOPPED)
	public void stop() {
		callback(STOPPED);
	}
	
	@On(DESTROYED)
	public void destroy() {
		callback(DESTROYED);
	}
	
	private void callback(Lifecycle lifecycle) {
		manager.callback(exposer, lifecycle);
		manager.callback(dispatcher, lifecycle);
	}
	
	/**
	 * Indicates that the specified service definition can be exposed or not.
	 * The request is forwarded to exposer directly.
	 * 
	 * @param service The service definition that deployer configured.
	 * @see Exposer
	 */
	public boolean expose(Service service) {
		return exposer.expose(service);
	}
	
	/**
	 * Detects the frontend that requires service to this {@code Bus} instance.
	 * 
	 * @param frontend The frontend class that requires service.
	 */
	public <F> void detect(Class<F> frontend) {
		if (!detected.contains(frontend)) {
			detected.add(frontend);
			logger.debug("Frontend [" + frontend + "] detected");
		}
	}
	
	/**
	 * Returns the service proxy with specified service key.
	 * The service key is forwarded to {@code Exposer#getService(Key)}. If 
	 * exposer doesn't hold the service definition that related to specified 
	 * service key, an {@code Exception} will be thrown.
	 * 
	 * @param key The key for looking up the related service.
	 * @return The service proxy that related to specified key.
	 * @see Exposer
	 */
	public Object getServiceProxy(Key<?> key) {
		Service service = exposer.getService(key);
		Class<?>[] interfaces = {key.getType()};
		return newProxyInstance(interfaces, service, key);
	}
	
	/**
	 * Dispatches a service method invocation.
	 * This method is invoked by the service proxy, and this method invocation 
	 * is forwarded to {@code Dispatcher#dispatch(Service, Method, Object...)}.
	 * 
	 * @param service The definition of service that is invoked by frontend of 
	 * other services.
	 * @param instance The service instance that is invoked by frontend of other 
	 * services.
	 * @param method The service method that is invoked.
	 * @param args The arguments for the service method invocation.
	 * @return The result by executing the service method invocation. 
	 * @throws Throwable Thrown if this service method invocation failed.
	 * @see Dispatcher
	 */
	public <R> R dispatch(Service service, Object instance, Method method, Object... args) throws Throwable {
		return dispatcher.<R>dispatch(service, instance, method, args);
	}
	
	private Object newProxyInstance(Class<?>[] interfaces, Service service, Key<?> key) {
		ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
		Set<Property<?>> properties = new HashSet<Property<?>>();
		properties.add(new Property<Key<?>>("key", key));
		
		if (service.getScope() == Scope.DEPENDENT) {
			properties.add(new Property<Object>("instance", exposer.getService(service)));
		}
		
		InvocationHandler invocationHandler = kernel.assemble(ServiceInvocationHandler.class).with(properties).getInstance();
		return Proxy.newProxyInstance(classLoader, interfaces, invocationHandler);
	}

	public Exposer getExposer() {
		return exposer;
	}

	public Dispatcher getDispatcher() {
		return dispatcher;
	}

	public void setKernel(Kernel kernel) {
		this.kernel = kernel;
	}

	public void setManager(Manager manager) {
		this.manager = manager;
	}
	
}
