/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.bus;

import static org.eiichiro.jazzmaster.management.Lifecycle.CONFIGURED;
import static org.eiichiro.jazzmaster.management.Lifecycle.LOADED;
import static org.eiichiro.jazzmaster.management.Lifecycle.STARTED;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eiichiro.jazzmaster.deployment.Scanner;
import org.eiichiro.jazzmaster.deployment.Stage;
import org.eiichiro.jazzmaster.kernel.Binding;
import org.eiichiro.jazzmaster.kernel.Injected;
import org.eiichiro.jazzmaster.kernel.Invoker;
import org.eiichiro.jazzmaster.kernel.Kernel;
import org.eiichiro.jazzmaster.kernel.Property;
import org.eiichiro.jazzmaster.management.Lifecycle;
import org.eiichiro.jazzmaster.management.Manager;
import org.eiichiro.jazzmaster.management.On;
import org.eiichiro.jazzmaster.repository.Key;
import org.eiichiro.jazzmaster.service.Consumes;
import org.eiichiro.jazzmaster.service.Refers;
import org.eiichiro.jazzmaster.service.Service;
import org.eiichiro.reverb.lang.UncheckedException;
import org.eiichiro.reverb.reflection.Invocation;
import org.eiichiro.reverb.time.Stopwatch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code Dispatcher} the executor of service method and service interceptor.
 * This is a built-in component that is managed in the {@code Bus} as singleton.
 * When frontend or service that is a consumer of other service invokes a 
 * service (in fact, it's a service proxy) method, the encoded invocation is 
 * send to bus. Then, bus forwards its invocation to {@code Dispatcher} instance 
 * and invokes {@code #dispatch(Service, Method, Object...)} method. At the 
 * first, dispatcher looks up service instance from service repository with 
 * specified service definition as a service instance key. If service instance 
 * returned from repository, dispatcher constructs the {@code Invoker} stack 
 * from {@code Kernel} with service instance, invoked service method, parameters 
 * and matched interceptors. After invoker stack constructed, dispatcher 
 * executes invoker and waits its result. In development stage or test stage, 
 * dispatcher indicates the performance about service method invocation.
 * 
 * @see Kernel
 * @see Invoker
 * @see Bus
 * @see Interceptor
 * @see Interception
 * @see Stage
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
@SuppressWarnings("deprecation")
public class Dispatcher {

	private Logger logger = LoggerFactory.getLogger(Dispatcher.class);
	
	@Injected private Scanner scanner;
	
	@Injected private Kernel kernel;
	
	@Injected private Stage stage;
	
	@Injected private Manager manager;
	
	@Injected private Bus bus;
	
	private Set<Object> interceptors = new HashSet<Object>();
	
	/**
	 * Constructs {@code Interceptor}s.
	 * This method is called by manager when this {@code Dispatcher} instance is 
	 * started by {@code Bus}.
	 * 
	 * @see Bus
	 * @see Manager
	 * @see Lifecycle
	 */
	@On(STARTED)
	public void start() {
		try {
			for (Class<?> clazz : scanner.getInterceptors()) {
				if (!Modifier.isPublic(clazz.getModifiers())) {
					logger.debug("Non-public interceptor [" + clazz.getName()
							+ "] is not loaded");
				} else if (Modifier.isAbstract(clazz.getModifiers())) {
					logger.debug("Abstract interceptor [" + clazz.getName()
							+ "] is not loaded");
				} else {
					manager.setCallback(clazz);
					manager.callback(clazz, LOADED);
					logger.debug("Interceptor [" + clazz.getName() + "] loaded");
					Object interceptor = kernel.assemble(clazz).getInstance();
					Set<Property<?>> services = new HashSet<Property<?>>();
					
					for (Property<?> property : kernel.disassemble(interceptor).getProperties()) {
						Set<Annotation> annotations = property.getAnnotations();
						boolean consumes = false;
						
						for (Annotation annotation : annotations) {
							Class<? extends Annotation> type = annotation.annotationType();
							
							if (type.equals(Refers.class) || type.equals(Consumes.class)) {
								consumes = true;
								break;
							}
						}
						
						if (consumes) {
							Set<Annotation> bindings = getBindings(annotations);
							Object service = bus.getServiceProxy(newKey(
									property.getType(), bindings));
							
							if (service != null) {
								services.add(newProperty(property.getName(),
										property.getType(), service));
							}
						}
					}
					
					kernel.assemble(interceptor).with(services).getInstance();
					manager.setCallback(interceptor);
					manager.callback(interceptor, CONFIGURED);
					interceptors.add(interceptor);
					manager.callback(interceptor, STARTED);
				}
			}
			
		} catch (Exception e) {
			throw new UncheckedException(e);
		}
	}
	
	/**
	 * Executes a service method invocation.
	 * At the first, this method gets {@code Interception} annotation from 
	 * invoked service method and specify the interceptors corresponding to them. 
	 * The next step, constructs the invocation stack with the service instance, 
	 * invoked method and interceptors. At the last, invokes the invocation stack 
	 * one by one, and returns the result to the client.
	 * 
	 * @param service The definition of service that is invoked by frontend or 
	 * other services.
	 * @param instance The service instance that is invoked by frontend or other 
	 * services.
	 * @param method The service method that is invoked.
	 * @param args The arguments for the service method invocation.
	 * @return The result by executing the service method invocation.
	 * @throws Throwable Thrown if this service method invocation failed.
	 * @see Invoker
	 * @see Interceptor
	 * @see Interception
	 */
	@SuppressWarnings("unchecked")
	public <R> R dispatch(Service service, Object instance, Method method,
			Object... args) throws Throwable {
		List<Object> interceptors = new ArrayList<Object>();
		
		if (args == null) {
			args = new Object[0];
		}
		
		for (Class<?> clazz : service.getInterceptors(method)) {
			for (Object interceptor : this.interceptors) {
				if (clazz.equals(interceptor.getClass())) {
					interceptors.add(interceptor);
				}
			}
		}
		
		Set<Annotation> interceptions = new HashSet<Annotation>();
		
		for (Annotation annotation : method.getAnnotations()) {
			if (annotation.annotationType().getAnnotation(Interception.class) != null) {
				interceptions.add(annotation);
			}
		}
		
		for (Annotation interception : interceptions) {
			for (Object interceptor : this.interceptors) {
				for (Annotation annotation : interceptor.getClass().getAnnotations()) {
					if (interception.equals(annotation)) {
						if (!interceptors.contains(interceptor)) {
							interceptors.add(interceptor);
						}
					}
				}
			}
		}
		
		List<Invoker> before = new ArrayList<Invoker>();
		List<Invoker> after = new ArrayList<Invoker>();
		List<Invoker> around = new ArrayList<Invoker>();
		List<Invoker> throwing = new ArrayList<Invoker>();
		
		for (Object interceptor : interceptors) {
			for (Method m : interceptor.getClass().getMethods()) {
				if (m.getAnnotation(Before.class) != null) {
					before.add(kernel.invoke(m).on(interceptor));
				} else if (m.getAnnotation(After.class) != null) {
					after.add(kernel.invoke(m).on(interceptor));
				} else if (m.getAnnotation(Throwing.class) != null) {
					throwing.add(kernel.invoke(m).on(interceptor));
				} else if (m.getAnnotation(Around.class) != null) {
					Invocation<?> invocation;
					
					if (around.size() == 0) {
						invocation = kernel.invoke(method).on(instance).with(
								Arrays.asList(args)).getInvocation();
					} else {
						Invoker invoker = around.get(around.size() - 1);
						Method getResult = Invoker.class.getMethod("getResult");
						invocation = kernel.invoke(getResult).on(invoker).getInvocation();
					}
					
					List<Object> arguments = new ArrayList<Object>(1);
					arguments.add(invocation);
					around.add(kernel.invoke(m).on(interceptor).with(arguments));
				}
			}
		}
		
		Stopwatch stopwatch = null;
		boolean production = (stage == Stage.PRODUCTION);
		
		if (!production) {
			stopwatch = new Stopwatch();
			stopwatch.start();
		}
		
		for (Invoker invoker : before) {
			invoker.getResult();
		}
		
		Object result = null;
		
		try {
			if (around.size() == 0) {
				result = kernel.invoke(method).on(instance).with(
						Arrays.asList(args)).getResult();
			} else {
				result = around.get(around.size() - 1).getResult();
			}
			
		} catch (Throwable t) {
			for (Invoker invoker : throwing) {
				List<Object> arguments = new ArrayList<Object>(1);
				arguments.add(t);
				invoker.with(arguments).getResult();
			}
			
			throw t;
		}
		
		for (Invoker invoker : after) {
			invoker.getResult();
		}
		
		if (!production) {
			stopwatch.stop();
			logger.trace("Service method [" + method
					+ "] has been invoked in [" + stopwatch.toString() + "]");
		}
		
		return (R) result; 
	}
	
	private Set<Annotation> getBindings(Set<Annotation> annotations) {
		Set<Annotation> bindings = new HashSet<Annotation>();
		
		for (Annotation annotation : annotations) {
			if (annotation.annotationType().getAnnotation(Binding.class) != null) {
				bindings.add(annotation);
			}
		}
		
		return bindings;
	}
	
	private <T> Key<T> newKey(Class<T> type, Set<Annotation> bindings) {
		return new Key<T>(type, bindings);
	}
	
	@SuppressWarnings("unchecked")
	private <T> Property<T> newProperty(String name, Class<T> type, Object value) {
		return new Property<T>(name, type, (T) value);
	}

	public Set<Object> getInterceptors() {
		return interceptors;
	}

	public void setKernel(Kernel kernel) {
		this.kernel = kernel;
	}

	public void setStage(Stage stage) {
		this.stage = stage;
	}

	public void setManager(Manager manager) {
		this.manager = manager;
	}

	public void setBus(Bus bus) {
		this.bus = bus;
	}

	public void setScanner(Scanner scanner) {
		this.scanner = scanner;
	}
	
}
