/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.bus;

import java.io.Serializable;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;

import org.eiichiro.jazzmaster.Application;
import org.eiichiro.jazzmaster.frontend.Injection;
import org.eiichiro.jazzmaster.kernel.Injected;
import org.eiichiro.jazzmaster.repository.Key;
import org.eiichiro.jazzmaster.service.Service;

/**
 * {@code ServiceInvocationHandler} is the implementation of 
 * {@code InvocationHandler} for service proxy instance.
 * In usually, you may not invoke service instance directly in Jazzmaster SOA. 
 * In Jazzmaster SOA, service declarations on application frontend and 
 * service implementation are all service proxy. {@code Bus} creates a service 
 * proxy with this class instance and {@code Injection} injects it into 
 * application frontends or the other services.
 *
 * @see Bus
 * @see Injection
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class ServiceInvocationHandler implements InvocationHandler, Serializable {

	private static final long serialVersionUID = -3465587170955568225L;

	@Injected private Key<?> key;
	
	@Injected private Object instance;
	
	/**
	 * Processes service proxy invocation.
	 * If {@code ServiceInvocationHandler#instance} is {@code null}, this method 
	 * requires the service instance from {@code Exposer} with service 
	 * definition and invokes {@code Bus#dispatch(Service, Object, Method, Object...)}. 
	 * If {@code ServiceInvocationHandler#instance} is not {@code null}, this 
	 * method invokes it with {@code ServiceInvocationHandler#instance}. It is 
	 * just the way of service scope control. If service scope is NOT 
	 * {@code Scope#DEPENDENT}, service scope is controlled in {@code Manager}, 
	 * so that service instance needs to be required from {@code Exposer}. If 
	 * service scope is {@code Scope#DEPENDENT} (default scope), the scope needs 
	 * to be controlled on the scope that application frontend or other service 
	 * that depends on this service is controlled on. {@code ServiceInvocationHandler} 
	 * and service proxy are instantiated for every service reference declaration 
	 * and injected into them. Therefore, service proxy's lifetime is the same 
	 * with the one of application frontend or the other service that depends on 
	 * it. Because of setting service instance in instance field, Jazzmaster 
	 * controls the dependent scope pretty well.
	 * 
	 * @see java.lang.reflect.InvocationHandler#invoke(java.lang.Object, java.lang.reflect.Method, java.lang.Object[])
	 * @see Bus
	 * @see Dispatcher
	 */
	public Object invoke(Object proxy, Method method, Object[] args)
			throws Throwable {
		Bus bus = Application.getManager().getBus();
		Exposer exposer = bus.getExposer();
		Service service = exposer.getService(key);
		Object instance = (this.instance != null) ? this.instance : exposer.getService(service);
		return bus.dispatch(service, instance, method, args);
	}
	
	public void setKey(Key<?> key) {
		this.key = key;
	}
	
	public void setInstance(Object instance) {
		this.instance = instance;
	}

}
