/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.deployment;

import java.io.IOException;
import java.util.HashSet;
import java.util.Set;

import org.eiichiro.jazzmaster.bus.Interceptor;
import org.eiichiro.jazzmaster.kernel.Injected;
import org.eiichiro.jazzmaster.service.Implementation;
import org.eiichiro.jazzmaster.service.Service;
import org.eiichiro.reverb.lang.ClassResolver;
import org.eiichiro.reverb.lang.UncheckedException;

/**
 * {@code ClasspathScanner} is a built-in implementation of {@code Scanner} 
 * interface. This implementation loads service definition classes and service 
 * implementation classes from classpath. This class uses {@link ClassResolver} 
 * internally.
 *
 * @see Deployer
 * @see org.eiichiro.reverb.lang.ClassResolver
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class ClasspathScanner implements Scanner {

	@Injected private ClassResolver classResolver;
	
	private Set<Class<?>> implementations = new HashSet<Class<?>>();
	
	private Set<Class<? extends Service>> services = new HashSet<Class<? extends Service>>();
	
	private Set<Class<?>> interceptors = new HashSet<Class<?>>();
	
	/**
	 * Returns service implementation classes on the classpath.
	 * 
	 * @return Service implementation classes on the classpath.
	 * @see org.eiichiro.jazzmaster.deployment.Scanner#getImplementations()
	 */
	public Set<Class<?>> getImplementations() {
		return implementations;
	}

	/**
	 * Returns service definition classes on the classpath.
	 * 
	 * @return Service definition classes on the classpath.
	 * @see org.eiichiro.jazzmaster.deployment.Scanner#getServices()
	 */
	public Set<Class<? extends Service>> getServices() {
		return services;
	}

	/**
	 * Scans service definition classes and service implementation classes from 
	 * classpath.
	 * A service definition class is the class that extends 
	 * {@code org.eiichiro.jazzmaster.service.Service}. A service implementation 
	 * class is the class that annotated 
	 * {@code org.eiichiro.jazzmaster.service.Implementation}. This method uses 
	 * {@code org.eiichiro.reverb.lang.ClassResolver} class with 
	 * {@code ClassLoader} set on the current thread for scanning. Therefore, 
	 * the classes are loaded from the classpath that the current 
	 * {@code ClassLoader} manages. If the {@code ClassLoader} on the current 
	 * thread has been changed by others before this method invoked, you may 
	 * receive an unexpected result.
	 * 
	 * @see Service
	 * @see Implementation
	 * @see org.eiichiro.jazzmaster.deployment.Scanner#scan()
	 */
	public void scan() {
		try {
			for (Class<? extends Service> service : classResolver
					.resolveBySuperclass(Service.class)) {
				services.add(service);
			}
			
			for (Class<?> implementation : classResolver
					.resolveByAnnotation(Implementation.class)) {
				implementations.add(implementation);
			}
			
			for (Class<?> interceptor : classResolver
					.resolveByAnnotation(Interceptor.class)) {
				interceptors.add(interceptor);
			}
			
		} catch (IOException e) {
			throw new UncheckedException(e);
		} catch (ClassNotFoundException e) {
			throw new UncheckedException(e);
		}
	}

	public void setClassResolver(ClassResolver classResolver) {
		this.classResolver = classResolver;
	}

	@Override
	public Set<Class<?>> getInterceptors() {
		return interceptors;
	}

}
