/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.deployment;

import static org.eiichiro.jazzmaster.management.Lifecycle.CONFIGURED;
import static org.eiichiro.jazzmaster.management.Lifecycle.DESTROYED;
import static org.eiichiro.jazzmaster.management.Lifecycle.LOADED;
import static org.eiichiro.jazzmaster.management.Lifecycle.STARTED;
import static org.eiichiro.jazzmaster.management.Lifecycle.STOPPED;
import static org.eiichiro.jazzmaster.management.Scope.REQUEST;
import static org.eiichiro.jazzmaster.management.Scope.SESSION;

import java.lang.reflect.Modifier;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;

import org.eiichiro.jazzmaster.Application;
import org.eiichiro.jazzmaster.bus.Bus;
import org.eiichiro.jazzmaster.kernel.Injected;
import org.eiichiro.jazzmaster.kernel.Kernel;
import org.eiichiro.jazzmaster.kernel.Property;
import org.eiichiro.jazzmaster.management.Configuration;
import org.eiichiro.jazzmaster.management.Lifecycle;
import org.eiichiro.jazzmaster.management.Manager;
import org.eiichiro.jazzmaster.management.Scope;
import org.eiichiro.jazzmaster.management.Scoped;
import org.eiichiro.jazzmaster.repository.Repository;
import org.eiichiro.jazzmaster.service.ConfigurationException;
import org.eiichiro.jazzmaster.service.Implementation;
import org.eiichiro.jazzmaster.service.Service;
import org.eiichiro.reverb.lang.ClassResolver;
import org.eiichiro.reverb.system.Environment;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code Deployer} deploys (loads, configures and starts) and undeploys 
 * Jazzmaster SOA components.
 * In Jazzmaster bootstrap, this class loads, configures and starts 
 * {@code Repository}, then loads, configures and starts {@code Bus}. After them, 
 * this class loads service definition using {@code Scanner}, configures and 
 * starts them. When SOA components are loaded, configured and started, deployer 
 * calls their lifecycle callback methods forwarding it to 
 * {@code Manager#callback(Object, org.eiichiro.jazzmaster.management.Lifecycle)}.
 * In Jazzmaster shutdown, this class stops and destroys {@code Scanner}s. The 
 * other SOA components (for example, {@code Repository}, {@code Bus}, 
 * {@code Service} and so on) are managed in {@code Manager}. So these component 
 * are stopped and destroyed immediately by {@code Manager}.
 *
 * @see Repository
 * @see Bus
 * @see Service
 * @see Scanner
 * @see Manager
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class Deployer {

	private Logger logger = LoggerFactory.getLogger(Deployer.class);
	
	@Injected private ClassResolver classResolver;
	
	@Injected private Kernel kernel;
	
	@Injected private Stage stage;
	
	@Injected private Configuration configuration;
	
	@Injected private Manager manager;
	
	private Scanner scanner;
	
	private Repository repository;
	
	private Bus bus;
	
	public static final String DEPLOYMENT_URL_PATTERN = "org.eiichiro.jazzmaster.deployment.url.pattern";
	
	public static final String SCANNER_IMPLEMENTATION = "org.eiichiro.jazzmaster.deployment.scanner.implementation";
	
	/**
	 * Deploys Jazzmaster SOA components.
	 * Deployer loads, configures, starts {@link Repository}, {@link Bus} and 
	 * {@link Service} (service definition) at bootstrap according to the 
	 * following details: 
	 * <ol>
	 * <li>
	 * Deployer loads {@code Repository.class} and calls back 
	 * {@code Lifecycle#LOADED} lifecycle method by using {@code Manager} and 
	 * assembles {@code Repository} with {@code ClassResolver}, {@code Kernel}, 
	 * {@code Stage}, {@code Configuration}, {@code Manager} and {@code Deployer} 
	 * (this instance). Then, calls back {@code Lifecycle#CONFIGURED} lifecycle 
	 * method. 
	 * </li>
	 * <li>
	 * In the same way of {@code Repository}, this method loads and configures 
	 * {@code Bus}. After {@code Bus} configured, reassembles {@code Repository} 
	 * with {@code Bus}.
	 * </li>
	 * <li>
	 * After {@code Bus} and {@code Repository} configured, deployer loads 
	 * {@code Scanner} implementation classes from classpath (using 
	 * {@code ClassResolver}). If loaded scanner is {@code public} and not 
	 * {@code abstract}, this method calls back {@code Lifecycle#LOADED} 
	 * lifecycle method, assembles it with {@code Repository}, {@code Bus}, 
	 * {@code ClassResolver}, {@code Kernel}, {@code Stage}, {@code Configuration}, 
	 * {@code Manager} and {@code Deployer} (these are "build-in components") 
	 * and calls back {@code Lifecycle#CONFIGURED}.
	 * </li>
	 * <li>
	 * Calls back {@code Lifecycle#STARTED} lifecycle method for each scanner 
	 * and Invokes {@code Scanner#scan()}. After all scanner started, deployer 
	 * gets all service definition classes and service implementation classes, 
	 * however non-public classes and abstract classes are not loaded. All 
	 * loaded classes are called their {@code Lifecycle#LOADED} callback method.
	 * </li>
	 * <li>
	 * Service definition classes are assembled with built-in components and 
	 * called back their {@code Lifecycle#CONFIGURED} method. Service 
	 * implementation classes are assembled as a anonymous service definition 
	 * instance and called back their {@code Lifecycle#CONFIGURED} method.
	 * </li>
	 * <li>
	 * All assembled service definition components are stored in 
	 * {@code Repository} and exposed by {@code Bus}. Then service definition 
	 * components is called {@code Lifecycle#STARTED} lifecycle callback method.
	 * </li>
	 * <li>
	 * Starts {@code Repository} and {@code Bus}, and calls back 
	 * {@code Lifecycle#STARTED} lifecycle method.
	 * </li>
	 * <li>
	 * {@code Manager} is reassembled with {@code Repository} and {@code Bus}. 
	 * Manager starts to manage deployed SOA components.
	 * </li>
	 * </ol>
	 * When any unexpected exception has occurred in deployment process, deployment 
	 * is stopped right away.
	 * 
	 * @see Manager#callback(Object, Lifecycle)
	 * @see Scanner
	 * @see Repository#store(Service)
	 * @see Bus#expose(Service)
	 */
	@SuppressWarnings("unchecked")
	public void deploy() {
		// Too long to read?
		logger.info("Starting deployment");
		logger.trace("Starting repository deployment");
		manager.setCallback(Repository.class);
		manager.callback(Repository.class, LOADED);
		Repository repository = kernel.assemble(Repository.class).getInstance();
		kernel.setGlobal(new Property<Repository>(Repository.class, repository));
		manager.setCallback(repository);
		manager.callback(repository, CONFIGURED);
		
		Class<? extends Scanner> scannerClass;
		Object value = configuration.get(SCANNER_IMPLEMENTATION);
		
		if (value instanceof Class<?>) {
			scannerClass = (Class<? extends Scanner>) value;
		} else {
			scannerClass = ClasspathScanner.class;
			Pattern pattern = null;
			Object object = configuration.get(DEPLOYMENT_URL_PATTERN);
			
			if (object instanceof Pattern) {
				pattern = Pattern.class.cast(object);
			}
			
			if (pattern != null) {
				List<URL> urls = new ArrayList<URL>();
				
				for (URL url : classResolver.getURLs()) {
					if (pattern.matcher(url.toString()).matches()) {
						urls.add(url);
					}
				}
				
				classResolver.setURLs(urls);
			}
			
			if (stage.compareTo(Stage.PRODUCTION) != 0) {
				List<URL> urls = classResolver.getURLs();
				StringBuilder stringBuilder = new StringBuilder();
				
				for (int i = 0; i < urls.size(); i++) {
					stringBuilder.append(urls.get(i));
					
					if (i < urls.size() - 1) {
						stringBuilder.append(Environment
								.getProperty("path.separator"));
					}
				}
				
				logger.debug("Deployment search path is [" + stringBuilder + "]");
			}
		}
		
		manager.setCallback(scannerClass);
		manager.callback(scannerClass, LOADED);
		logger.debug("Scanner [" + scannerClass.getName() + "] loaded");
		Scanner scanner = kernel.assemble(scannerClass).getInstance();
		manager.setCallback(scanner);
		manager.callback(scanner, CONFIGURED);
		kernel.setGlobal(new Property<Scanner>(Scanner.class, scanner));
		
		logger.trace("Starting bus deployment");
		manager.setCallback(Bus.class);
		manager.callback(Bus.class, LOADED);
		Bus bus = kernel.assemble(Bus.class).getInstance();
		kernel.setGlobal(new Property<Bus>(Bus.class, bus));
		kernel.assemble(repository).getInstance();
		manager.setCallback(bus);
		manager.callback(bus, CONFIGURED);
		
		logger.trace("Starting services deployment");
		Set<Class<? extends Service>> classes 
				= new HashSet<Class<? extends Service>>();
		Set<Class<?>> implementations = new HashSet<Class<?>>();
		Set<Class<?>> loaded = new HashSet<Class<?>>();
		
		manager.callback(scanner, STARTED);
		scanner.scan();
		
		for (Class<? extends Service> service : scanner.getServices()) {
			if (!Modifier.isPublic(service.getModifiers())) {
				logger.debug("Non-public service [" + service.getName()
						+ "] is not loaded");
			} else if (Modifier.isAbstract(service.getModifiers())) {
				logger.debug("Abstract service [" + service.getName()
						+ "] is not loaded");
			} else {
				classes.add(service);
			}
		}
		
		for (Class<?> implementation : scanner.getImplementations()) {
			if (!Modifier.isPublic(implementation.getModifiers())) {
				logger.debug("Non-public service implementation ["
						+ implementation.getName() + "] is not loaded");
			} else if (Modifier.isAbstract(implementation.getModifiers())) {
				logger.debug("Abstract service implementation ["
						+ implementation.getName() + "] is not loaded");
			} else if (implementation.getAnnotation(Implementation.class) == null) {
				logger.debug("Service implementation ["
						+ implementation.getName()
						+ "] must be annotated @"
						+ Implementation.class.getName());
			} else {
				implementations.add(implementation);
			}
		}
		
		Set<Service> services = new HashSet<Service>();
		
		for (Class<? extends Service> clazz : classes) {
			Staged staged = clazz.getAnnotation(Staged.class);
			
			if (staged != null) {
				if (Arrays.binarySearch(staged.value(), stage) < 0) {
					continue;
				}
			}
			
			manager.setCallback(clazz);
			manager.callback(clazz, LOADED);
			logger.debug("Service [" + clazz.getName() + "] loaded");
			Service service = kernel.assemble(clazz).getInstance();
			
			if (service != null) {
				if (!service.getStages().contains(stage)) {
					continue;
				}
				
				Class<?> implementation = service.getImplementation();
				
				if (implementation != null) {
					if (!loaded.contains(implementation)) {
						staged = implementation.getAnnotation(Staged.class);
						
						if (staged != null) {
							if (Arrays.binarySearch(staged.value(), stage) < 0) {
								continue;
							}
						}
						
						manager.setCallback(implementation);
						manager.setCallback(implementation, LOADED, service.getCallbacks(LOADED));
						manager.callback(implementation, LOADED);
						loaded.add(implementation);
					}
					
					manager.setCallback(service);
					manager.callback(service, CONFIGURED);
					services.add(service);
				}
				
			} else {
				logger.debug("Service [" + clazz.getName()
						+ "] cannot be configured");
			}
		}
		
		for (final Class<?> clazz : implementations) {
			final Staged staged = clazz.getAnnotation(Staged.class);
			
			if (staged != null) {
				if (Arrays.binarySearch(staged.value(), stage) < 0) {
					continue;
				}
			}
			
			if (!loaded.contains(clazz)) {
				manager.setCallback(clazz);
				manager.callback(clazz, LOADED);
				loaded.add(clazz);
				logger.debug("Service implementation ["
						+ clazz.getName() + "] loaded");
			}
			
			final Scoped scoped = clazz.getAnnotation(Scoped.class);
			Service service = new ServiceImpl(clazz, scoped, staged);
			service = kernel.assemble(service).getInstance();
			manager.setCallback(service);
			manager.callback(service, CONFIGURED);
			services.add(service);
		}
		
		for (Service service : services) {
			repository.store(service);
			
			if (bus.expose(service)) {
				manager.callback(service, STARTED);
				logger.info("Service [" + service.getName() + "] has been deployed");
			} else {
				logger.warn("Service [" + service.getName() + "] has not been deployed");
			}
		}
		
		logger.info("Services deployment completed");
		manager.callback(repository, STARTED);
		logger.info("Repository deployment completed");
		manager.callback(bus, STARTED);
		logger.info("Bus deployment completed");
		this.scanner = scanner;
		this.repository = repository;
		this.bus = bus;
		kernel.assemble(manager).with(new Property<Repository>(repository)).getInstance();
		kernel.assemble(manager).with(new Property<Bus>(bus)).getInstance();
		logger.info("Deployment completed on [" + stage + "] stage");
	}
	
	private static class ServiceImpl extends Service {

		private final Class<?> implementation;
		
		private final Scoped scoped;
		
		private final Staged staged;
		
		private ServiceImpl(Class<?> implementation, Scoped scoped, Staged staged) {
			this.implementation = implementation;
			this.scoped = scoped;
			this.staged = staged;
		}
		
		@Override
		public Class<?> getImplementation() {
			return implementation;
		}
		
		@Override
		public String getName() {
			return implementation.getName();
		}
		
		@Override
		public Scope getScope() {
			if (scoped != null) {
				Scope scope = scoped.value();
				
				if (!Application.isWeb() && (scope == REQUEST || scope == SESSION)) {
					throw new ConfigurationException("Service ["
							+ getName() + "is misconfigured: @Scoped ["
							+ scoped + "] cannot be available");
				}
				
				return scope;
			} else {
				return super.getScope();
			}
		}
		
		@Override
		public Set<Stage> getStages() {
			if (staged != null) {
				Set<Stage> stages = new HashSet<Stage>(staged.value().length);
				
				for (Stage stage : staged.value()) {
					stages.add(stage);
				}
				
				return stages;
			} else {
				return super.getStages();
			}
		}
		
	}
	
	/**
	 * Undeploys Jazzmaster SOA components.
	 * Stops and destroys all scanners using 
	 * {@code Manager#callback(Object, Lifecycle)}.
	 */
	public void undeploy() {
		logger.trace("Starting undeployment");
		manager.callback(scanner, STOPPED);
		manager.callback(scanner, DESTROYED);
		logger.info("Undeployment completed");
	}
	
	public Repository getRepository() {
		return repository;
	}

	public Bus getBus() {
		return bus;
	}

	public void setClassResolver(ClassResolver classResolver) {
		this.classResolver = classResolver;
	}

	public void setKernel(Kernel kernel) {
		this.kernel = kernel;
	}

	public void setStage(Stage stage) {
		this.stage = stage;
	}

	public void setManager(Manager manager) {
		this.manager = manager;
	}
	
	public void setConfiguration(Configuration configuration) {
		this.configuration = configuration;
	}
	
}
