/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.deployment;

/**
 * {@code Stage} is the Jazzmaster component deployment stage.
 * Bootstrapping your application specifying any stage, you can switch the 
 * deployed components depending on the specified stage with {@code Staged} 
 * annotation. See {@link Staged}.
 * And the stage may be populated into your custom components by Jazzmaster 
 * dependency injection. So that, you can configure your components with the 
 * current stage.
 * <pre>
 * package jazzmaster.hello;
 * 
 * import org.eiichiro.jazzmaster.deployment.Stage;
 * import org.eiichiro.jazzmaster.kernel.Injected;
 * import org.eiichiro.jazzmaster.management.Lifecycle;
 * import org.eiichiro.jazzmaster.management.On;
 * import org.eiichiro.jazzmaster.service.Implementation;
 * 
 * {@code @Implementation}
 * public class HelloImpl implements Hello {
 * 
 *     {@code @Injected} private Stage stage;
 * 	
 *     {@code @On(Lifecycle.STARTED)}
 *     public void start() {
 *         System.out.println();
 *         System.out.println("HelloImpl started; Stage is [" + stage + "]");
 *         System.out.println();
 *     }
 * 
 *     public void setStage(Stage stage) {
 *         this.stage = stage;
 *     }
 * 
 * }
 * </pre>
 * Application bootstrap is: 
 * <pre>
 * package jazzmaster.hello;
 * 
 * import org.eiichiro.jazzmaster.Application;
 * 
 * public class HelloMain {
 * 
 *     public static void main(String[] args) {
 *         Application.bootstrap(Stage.TEST);
 *         HelloFrontend helloFrontend = new HelloFrontend();
 *         HelloFrontend frontend = Application.getFrontend(HelloFrontend.class);
 *         helloFrontend.sayHello();
 *         frontend.sayHello();
 *         Application.shutdown();
 *     }
 * 
 * }
 * </pre>
 *
 * @see Staged
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public enum Stage {

	/** The test stage. */
	TEST, 
	
	/** The development stage. */
	DEVELOPMENT, 
	
	/** The production stage. */
	PRODUCTION, 
	
	/** The emergency stage. */
	EMERGENCY
	
}
