/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.kernel;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import org.eiichiro.reverb.reflection.Invocation;
import org.eiichiro.reverb.reflection.MethodInvocation;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code Invoker} constructs and executes an method invocation.
 * This class is used by {@code Kernel} for callback method invocation. If you 
 * want to use this class to callback method, you should control this class 
 * through {@code Kernel} with the 
 * <a href="http://www.martinfowler.com/bliki/FluentInterface.html">fluent 
 * interface</a> (like {@link Assembler}). For example: 
 * <pre>
 * package jazzmaster.hello;
 * 
 * import java.lang.reflect.Method;
 * import java.util.ArrayList;
 * import java.util.List;
 * 
 * import org.eiichiro.jazzmaster.kernel.Injected;
 * import org.eiichiro.jazzmaster.kernel.Kernel;
 * import org.eiichiro.jazzmaster.kernel.Property;
 * import org.eiichiro.jazzmaster.service.Implementation;
 * 
 * {@code @Implementation}
 * public class HelloImpl implements Hello {
 * 
 *     {@code @Injected} private Kernel kernel;
 * 	
 *     public Message sayHello() {
 *         try {
 *             // Constructs the method to invoke.
 *             Method method = Messenger.class.getMethod("getMessage", String.class);
 *             // Constructs the instance to invoke.
 *             Messenger messenger = new Messenger();
 *             // Constructs the arguments passed to invocation.
 *             List&lt;Object&gt; arguments = new ArrayList&lt;Object&gt;();
 *             arguments.add("Eiichiro!");
 *             
 *             // Invoke the method. If you want to use kernel independently 
 *             // from Jazzmaster framework, you need to instantiate Kernel 
 *             // directly.
 *             String message = kernel.invoke(method).with(arguments).on(messenger).getResult();
 *             System.out.println(message);
 *         } catch (Throwable t) {
 *             System.out.println("Error");
 *         }
 *     
 *         return kernel.assemble(Message.class).with(new Property&lt;String&gt;("Hello, Jazzmaster!")).getInstance();
 *     }
 * 
 *     public void setKernel(Kernel kernel) {
 *         this.kernel = kernel;
 *     }
 * 
 * }
 * </pre>
 * {@code Messenger} class is: 
 * <pre>
 * package jazzmaster.hello;
 * 
 * public class Messenger {
 * 
 *     public String getMessage(String string) {
 *         return "Hello, " + string;
 *     }
 * 
 * }
 * </pre>
 * You might think it is too tiresome, but this mechanism will provide the 
 * flexibility of method invocation to your programming model. Using this class 
 * for method invocation, you can invoke a method any number of times at any time. 
 * In Jazzmaster framework, this class is used for dynamic invocation of 
 * callback method and building of method invocation chain (by {@code Invoker} 
 * invokes {@code Invoker} method).
 *
 * @see Kernel
 * @see Method
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class Invoker {

	private Logger logger = LoggerFactory.getLogger(Invoker.class);
	
	private final Method method;
	
	private Object target;
	
	private List<Object> arguments = new ArrayList<Object>();
	
	private Object result;
	
	private Throwable thrown;
	
	private boolean invoked = false;
	
	/**
	 * Constructs a new {@code Invoker} instance with the specified 
	 * {@code Method}.
	 * 
	 * @param method The method to be invoked.
	 */
	public Invoker(Method method) {
		if (method == null) {
			throw new IllegalArgumentException("'method' must not be ["
					+ method + "]");
		}
		
		this.method = method;
	}
	
	/**
	 * Invokes the specified method on the specified instance with the specified 
	 * arguments and holds the invocation result.
	 * If the instance is not specified, this method invokes the class method. 
	 * The other, the instance is specified, this method invokes the instance 
	 * method. Once this method has invoked, the invocation result (if any 
	 * {@code Throwable} is thrown, this is {@code Throwable}) is cached, so 
	 * that {@code Invoker#getResult()} always returns the same result unless 
	 * target instance or arguments are reset.
	 * 
	 * @see Method#invoke(Object, Object...)
	 * @throws Throwable If invoked method throws any exception or fails to 
	 * invoke the method.
	 */
	public void invoke() throws Throwable {
		try {
			result = method.invoke(target, arguments.toArray());
		} catch (InvocationTargetException e) {
			thrown = e.getTargetException();
			logger.debug(thrown.getMessage(), thrown);
			throw thrown;
		} catch (Throwable t) {
			thrown = t;
			logger.debug(t.getMessage(), t);
			throw t;
		} finally {
			invoked = true;
		}
	}
	
	/**
	 * Sets the instance that a method is invoked.
	 * If the instance is {@code null} or not specified, this {@code Invoker} 
	 * invokes the class method (static method).
	 * 
	 * @param target The target instance that a method is invoked.
	 * @return This {@code Invoker} instance.
	 */
	public Invoker on(Object target) {
		this.target = target;
		invoked = false;
		return this;
	}
	
	/**
	 * Sets the arguments that a method is invoked with.
	 * 
	 * @param arguments The arguments that a method is invoked with.
	 * @return This {@code Invoker} instance.
	 */
	public Invoker with(List<Object> arguments) {
		if (arguments == null) {
			throw new IllegalArgumentException("'arguments' must not be ["
					+ arguments + "]");
		}
		
		this.arguments = arguments;
		invoked = false;
		return this;
	}
	
	/**
	 * Returns the invocation result.
	 * If any {@code Throwable} is thrown, this method throws the 
	 * {@code Throwable}. If {@code #invoke()} has not invoked yet, this method 
	 * invokes it and returns result or throws {@code Throwable}. This method is 
	 * often invoked at the end of fluent interface invocation flow. For 
	 * example: 
	 * <pre></pre>
	 * 
	 * @return The result that the method returns.
	 * @throws Throwable If the method throws any {@code Throwable}.
	 */
	@SuppressWarnings("unchecked")
	public <R> R getResult() throws Throwable {
		if (!invoked) {
			invoke();
			return this.<R>getResult();
		} else if (thrown != null) {
			throw thrown;
		} else {
			return (R) result;
		}
	}
	
	/**
	 * Returns the method {@code Invocation} representation that this 
	 * {@code Invoker} invokes. {@code Invocation} can be proceed at any time.
	 * 
	 * @see Invocation
	 * @see MethodInvocation
	 * @return the invocation
	 */
	public Invocation<?> getInvocation() {
		return newInvocation(method.getReturnType());
	}
	
	private <R> Invocation<R> newInvocation(Class<R> type) {
		return new MethodInvocation<R>(method, target, arguments.toArray());
	}
	
}
