/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.kernel;

import java.lang.reflect.Method;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code Kernel} is the core of Jazzmaster IoC implementation.
 * You can use the Jazzmaster IoC (dependency injection and callback invocation) 
 * mechanism through this class with the fluent interface programming. 
 * {@code Kernel} traces any IoC operation for debugging. You can use Jazzmaster 
 * IoC kernel independently from whole Jazzmaster SOA framework (can use 
 * {@code org.eiichiro.jazzmaster.kernel} package only). You had better read the 
 * sample code in {@code Assembler}, {@code Disassembler} and {@code Invoker}.
 *
 * @see Assembler
 * @see Disassembler
 * @see Invoker
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class Kernel {

	private Logger logger = LoggerFactory.getLogger(Kernel.class);
	
	private Set<Property<?>> globals = new CopyOnWriteArraySet<Property<?>>();
	
	/**
	 * Sets the global property.
	 * 
	 * @param globals The global property to set.
	 */
	public void setGlobal(Property<?> global) {
		logger.trace("Global property [" + global + "] is set");
		globals.add(global);
	}

	/**
	 * Returns the global properties.
	 * 
	 * @return The global properties.
	 */
	public Set<Property<?>> getGlobals() {
		return globals;
	}
	
	/**
	 * Constructs a new {@code Assembler} instance with the specified 
	 * {@code Class}.
	 * 
	 * @param clazz The {@code Class} to be assembled.
	 * @return The {@code Assembler} instance that is constructed with the 
	 * specified class.
	 */
	public <T> Assembler<T> assemble(Class<T> clazz) {
		logger.trace("Assembling class [" + clazz.getName() + "]");
		return new Assembler<T>(clazz).with(globals);
	}
	
	/**
	 * Constructs a new {@code Assembler} instance with the specified instance.
	 * 
	 * @param instance The instance to be assembled.
	 * @return The {@code Assembler} instance that is constructed with the 
	 * specified instance.
	 */
	public <T> Assembler<T> assemble(T instance) {
		logger.trace("Assembling instance [" + instance + "]");
		return new Assembler<T>(instance).with(globals);
	}
	
	/**
	 * Constructs a new {@code Disassembler} instance with the specified 
	 * {@code Class}.
	 * 
	 * @param clazz The {@code Class} to be disassembled.
	 * @return The {@code Disassembler} instance that is constructed with the 
	 * specified class.
	 */
	public <T> Disassembler<T> disassemble(Class<T> clazz) {
		logger.trace("Disassembling class [" + clazz.getName() + "]");
		return new Disassembler<T>(clazz);
	}
	
	/**
	 * Constructs a new {@code Disassembler} instance with the specified 
	 * instance.
	 * 
	 * @param instance The instance to be disassembled.
	 * @return The {@code Disassembler} instance that is constructed with the 
	 * specified instance.
	 */
	public <T> Disassembler<T> disassemble(T instance) {
		logger.trace("Disassembling instance [" + instance + "]");
		return new Disassembler<T>(instance);
	}
	
	/**
	 * Constructs a new {@code Invoker} instance with the specified 
	 * {@code Method}.
	 * 
	 * @param method The {@code Method} to be invoked.
	 * @return The {@code Invoker} instance that is constructed with the 
	 * specified method.
	 */
	public Invoker invoke(Method method) {
		logger.trace("Invoking method [" + method + "]");
		return new Invoker(method);
	}
	
}
