/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.kernel;

import java.lang.annotation.Annotation;
import java.util.HashSet;
import java.util.Set;

/**
 * {@code Property} is an instance field representation in Jazzmaster dependency 
 * injection. You can assemble the component which this instance value set in 
 * its field by passing this class instance to {@code Assembler}. Jazzmaster 
 * dependency injection usually <strong>typesafe</strong>. So, this instance's 
 * {@code #type} (if {@code #type} hasn't been specified, all the hierarchy of 
 * {@code #value}'s type) and annotations ({@code Binding} annotation) are used 
 * for specifying injected field. If {@code #name} has been specified, this 
 * instance matches the same name field (however, when the field's type doesn't 
 * match this type, this property isn't injected into it).
 * You can also disassemble and get the {@code Set} of {@code Property} which a 
 * component has, by specifying a instance to {@code Disassembler}. The property 
 * instance that disassembled from {@code Disassembler} has its {@code #name}, 
 * {@code #type}, {@code #value} and {@code #annotations}.
 *
 * @see Assembler
 * @see Disassembler
 * @see Binding
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class Property<T> {

	private String name;
	
	private Class<T> type;
	
	private final T value;
	
	private Set<Annotation> annotations = new HashSet<Annotation>();
	
	/**
	 * Constructs a new {@code Property} instance with the specified value.
	 * 
	 * @param value The value to be represented.
	 */
	public Property(T value) {
		if (value == null) {
			throw new IllegalArgumentException("'value' must not be [" + null
					+ "]");
		}
		
		this.value = value;
	}
	
	/**
	 * Constructs a new {@code Property} instance with the specified name and 
	 * value.
	 * 
	 * @param name The name to be represented.
	 * @param value The value to be represented.
	 */
	public Property(String name, T value) {
		if (value == null && (name == null || name.equals(""))) {
			throw new IllegalArgumentException("'name' must not be [" + name
					+ "] when 'value' is [" + value + "]");
		}
		
		this.name = name;
		this.value = value;
	}
	
	/**
	 * Constructs a new {@code Property} instance with the specified 
	 * {@code Class} and value. 
	 * 
	 * @param type The {@code Class} to be represented.
	 * @param value The value to be represented.
	 */
	public Property(Class<T> type, T value) {
		if (type == null && value == null) {
			throw new IllegalArgumentException("'type' must not be [" + type
					+ "] when 'value' is [" + value + "]");
		}
		
		this.type = type;
		this.value = value;
	}
	
	/**
	 * Constructs a new {@code Property} instance with the specified name, 
	 * {@code Class} and value.
	 * 
	 * @param name The name to be represented.
	 * @param type The {@code Class} to be represented.
	 * @param value The value to be represented.
	 */
	public Property(String name, Class<T> type, T value) {
		if (type == null && (name == null || name.equals("")) && value == null) {
			throw new IllegalArgumentException("'name' must not be [" + name
					+ "] when 'type' is [" + type + "] and 'value' is ["
					+ value + "]");
		}
		
		this.name = name;
		this.type = type;
		this.value = value;
	}
	
	/**
	 * Constructs a new {@code Property} instance with the specified value and 
	 * the {@code Set} of {@code Annotation}s.
	 * 
	 * @param value The value to be represented.
	 * @param annotations The {@code Set} of {@code Annotation} to be 
	 * represented.
	 */
	public Property(T value, Set<Annotation> annotations) {
		if (annotations == null) {
			throw new IllegalArgumentException("'annotations' must not be ["
					+ annotations + "]");
		} else {
			for (Annotation annotation : annotations) {
				if (annotation == null) {
					throw new IllegalArgumentException(
							"'annotations' must not contain [" + annotation
									+ "]");
				}
			}
		}
		
		if (value == null) {
			throw new IllegalArgumentException("'value' must not be [" + value
					+ "]");
		}
		
		this.value = value;
		this.annotations = annotations;
	}
	
	/**
	 * Constructs a new {@code Property} instance with the specified name, value 
	 * and the {@code Set} of {@code Annotation}.
	 * 
	 * @param name The name to be represented.
	 * @param value The value to be represented.
	 * @param annotations The {@code Set} of {@code Annotation} to be 
	 * represented.
	 */
	public Property(String name, T value, Set<Annotation> annotations) {
		if (annotations == null) {
			throw new IllegalArgumentException("'annotations' must not be ["
					+ null + "]");
		} else {
			for (Annotation annotation : annotations) {
				if (annotation == null) {
					throw new IllegalArgumentException(
							"'annotations' must not contain [" + annotation
									+ "]");
				}
			}
		}
		
		if (value == null && (name == null || name.equals(""))) {
			throw new IllegalArgumentException("'name' must not be [" + name
					+ "] when 'value' is [" + value + "]");
		}
		
		this.name = name;
		this.value = value;
		this.annotations = annotations;
	}
	
	/**
	 * Constructs a new {@code Property} instance with the specified {@code Class}, 
	 * value and the {@code Set} of {@code Annotation}.
	 * 
	 * @param type The {@code Class} to be represented.
	 * @param value The value to be represented.
	 * @param annotations The {@code Set} of {@code Annotation}.
	 */
	public Property(Class<T> type, T value, Set<Annotation> annotations) {
		if (annotations == null) {
			throw new IllegalArgumentException("'annotations' must not be ["
					+ null + "]");
		} else {
			for (Annotation annotation : annotations) {
				if (annotation == null) {
					throw new IllegalArgumentException(
							"'annotations' must not contain [" + annotation
									+ "]");
				}
			}
		}
		
		if (type == null && value == null) {
			throw new IllegalArgumentException("'type' must not be [" + type
					+ "] when 'value' is [" + value + "]");
		}
		
		this.type = type;
		this.value = value;
		this.annotations = annotations;
	}
	
	/**
	 * Constructs a new {@code Property} instance with the specified name, 
	 * {@code Class}, value and the {@code Set} of {@code Annotation}.
	 * 
	 * @param name The name to be represented.
	 * @param type The {@code Class} to be represented.
	 * @param value The value to be represented.
	 * @param annotations The {@code Set} of {@code Annotation}.
	 */
	public Property(String name, Class<T> type, T value, Set<Annotation> annotations) {
		if (annotations == null) {
			throw new IllegalArgumentException("'annotations' must not be ["
					+ null + "]");
		} else {
			for (Annotation annotation : annotations) {
				if (annotation == null) {
					throw new IllegalArgumentException(
							"'annotations' must not contain [" + annotation
									+ "]");
				}
			}
		}
		
		if ((name == null || name.equals("")) && type == null && value == null) {
			throw new IllegalArgumentException("'name' must not be [" + name
					+ "] when 'type' is [" + type + "] and 'value' is ["
					+ value + "]");
		}
		
		this.name = name;
		this.type = type;
		this.value = value;
		this.annotations = annotations;
	}
	
	/**
	 * @return the value
	 */
	public T getValue() {
		return value;
	}
	
	/**
	 * @return the name
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * @return the annotations
	 */
	public Set<Annotation> getAnnotations() {
		return annotations;
	}
	
	/**
	 * @return the type
	 */
	public Class<T> getType() {
		return type;
	}
	
}
