/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.management;

import static org.eiichiro.jazzmaster.management.Lifecycle.DESTROYED;
import static org.eiichiro.jazzmaster.management.Lifecycle.FAILED;
import static org.eiichiro.jazzmaster.management.Lifecycle.STOPPED;
import static org.eiichiro.jazzmaster.management.Scope.REQUEST;
import static org.eiichiro.jazzmaster.management.Scope.SESSION;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;

import org.apache.commons.lang.ClassUtils;
import org.eiichiro.jazzmaster.Application;
import org.eiichiro.jazzmaster.bus.Bus;
import org.eiichiro.jazzmaster.kernel.Injected;
import org.eiichiro.jazzmaster.kernel.Kernel;
import org.eiichiro.jazzmaster.repository.Repository;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * {@code Manager} provides scoped component's context management and component 
 * lifecycle management.
 * {@code Manager} manages all component {@code Scope}s and manages all scoped 
 * components by binding them into each scope's {@code Context}. And invokes 
 * component lifecycle callback method and manages component's current lifecycle.
 *
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class Manager {

	private Logger logger = LoggerFactory.getLogger(Manager.class);
	
	@Injected private Kernel kernel;
	
	@Injected private Repository repository;
	
	@Injected private Bus bus;
	
	private Map<Object, Lifecycle> lifecycles 
			= new WeakHashMap<Object, Lifecycle>();
	
	private Map<Object, Map<Lifecycle, Set<Method>>> callbacks 
			= new WeakHashMap<Object, Map<Lifecycle,Set<Method>>>();
	
	/**
	 * Starts component scope management.
	 * This method is invoked in bootstrap process and activates {@code Scope}s.
	 */
	public void manage() {
		logger.trace("Starting management");
		
		for (Scope scope : Scope.values()) {
			if (!Application.isWeb() && (scope == REQUEST || scope == SESSION)) {
				logger.warn("Web scope [" + scope + "] cannot be activated: "
						+ "The current application doesn't run in "
						+ "web application environment");
			} else {
				logger.info("Scope [" + scope + "] is activated");
			}
		}
		
		logger.info("Management started");
	}
	
	/**
	 * Stops component scope management.
	 * This method is invoked in shutdown process. At the first, {@code Manager} 
	 * stops all scoped component and stops all scopes. Next, this method 
	 * destroys all scoped component and destroy all scopes. At the last, this 
	 * method stops and destroys {@code Bus} and {@code Repository}.
	 */
	public void unmanage() {
		logger.trace("Stopping management");
		
		for (Scope scope : Scope.values()) {
			if (!Application.isWeb() && (scope == REQUEST || scope == SESSION)) {
				continue;
			}
			
			Context context = scope.getContext();
			
			for (Object key : context.keySet()) {
				Object object = context.get(key);
				
				if (object != null) {
					callback(object, STOPPED);
					logger.debug("Component [" + object + "] stopped");
				}
			}
			
			logger.debug("Scope [" + scope + "] stopped");
		}
		
		for (Scope scope : Scope.values()) {
			if (!Application.isWeb() && (scope == REQUEST || scope == SESSION)) {
				continue;
			}
			
			Context context = scope.getContext();
			
			for (Object key : context.keySet()) {
				Object object = context.get(key);
				
				if (object != null) {
					callback(object, DESTROYED);
					logger.debug("Component [" + object + "] destroyed");
				}
			}
			
			scope.getContext().clear();
			logger.debug("Scope [" + scope + "] destroyed");
		}
		
		callback(bus, STOPPED);
		callback(repository, STOPPED);
		callback(bus, DESTROYED);
		callback(repository, DESTROYED);
		logger.trace("Management stopped");
	}
	
	/**
	 * Sets lifecycle callback methods ({@code @On} annotated methods) declared 
	 * on the specified object of each lifecycle.
	 * 
	 * @param object The target component to be called back.
	 */
	public void setCallback(Object object) {
		for (Lifecycle lifecycle : Lifecycle.values()) {
			setCallback(object, lifecycle, new HashSet<Method>(0));
		}
	}
	
	/**
	 * Sets lifecycle callback methods corresponding to object and lifecycle.
	 * {@code @On} annotated method is registered automatically.
	 * 
	 * @param object The target component to be called back.
	 * @param lifecycle The lifecycle when the specified callbacks are invoked.
	 * @param callbacks The callback methods to be invoked on the specified 
	 * lifecycle.
	 */
	@SuppressWarnings("unchecked")
	public void setCallback(Object object, Lifecycle lifecycle, Set<Method> callbacks) {
		if (object == null) {
			throw new IllegalArgumentException("'target' must not be ["
					+ object + "]");
		} else if (lifecycle == null) {
			throw new IllegalArgumentException("'lifecycle' must not be ["
					+ lifecycle + "]");
		} else if (callbacks == null) {
			throw new IllegalArgumentException("'callbacks' must not be ["
					+ callbacks + "]");
		}
		
		List<Method> methods = new ArrayList<Method>();
		Class<?> clazz = (object instanceof Class<?>) ? Class.class.cast(object) : object.getClass();
		methods.addAll(Arrays.asList(clazz.getMethods()));
		
		for (Class<?> superclass : (List<Class<?>>) ClassUtils.getAllSuperclasses(clazz)) {
			for (Method method : superclass.getMethods()) {
				if (!methods.contains(method)) {
					methods.add(method);
				}
			}
		}
		
		for (Class<?> interfase : (List<Class<?>>) ClassUtils.getAllInterfaces(clazz)) {
			for (Method method : interfase.getMethods()) {
				if (!methods.contains(method)) {
					methods.add(method);
				}
			}
		}
		
		for (Method method : methods) {
			On on = method.getAnnotation(On.class);
			
			if ((on != null && Arrays.asList(on.value()).contains(lifecycle))
					|| callbacks.contains(method)) {
				Map<Lifecycle, Set<Method>> map = this.callbacks.get(object);
				
				if (map != null) {
					Set<Method> set = map.get(lifecycle);
					
					if (set != null) {
						if (!set.contains(method)) {
							set.add(method);
						}
						
					} else {
						Set<Method> s = new HashSet<Method>();
						s.add(method);
						map.put(lifecycle, s);
					}
					
				} else {
					Map<Lifecycle, Set<Method>> m = new HashMap<Lifecycle, Set<Method>>();
					Set<Method> set = new HashSet<Method>();
					set.add(method);
					m.put(lifecycle, set);
					this.callbacks.put(object, m);
				}
			}
		}
	}
	
	/**
	 * Invokes component lifecycle callback method.
	 * If lifecycle callback is failed, this method invokes 
	 * {@code Lifecycle#FAILED} lifecycle callback on the target component.
	 * 
	 * @param target The target component to be called back.
	 * @param lifecycle The current lifecycle of target component.
	 */
	public void callback(Object target, Lifecycle lifecycle) {
		if (target == null) {
			throw new IllegalArgumentException("'target' must not be ["
					+ target + "]");
		} else if (lifecycle == null) {
			throw new IllegalArgumentException("'lifecycle' must not be ["
					+ lifecycle + "]");
		}
		
		if (lifecycles.get(target) == lifecycle) {
			return;
		}
		
		Map<Lifecycle, Set<Method>> map = callbacks.get(target);
		
		if (map != null) {
			Set<Method> set = map.get(lifecycle);
			
			if (set != null) {
				for (Method method : set) {
					try {
						logger.trace("Calling back [" + method + "] on ["
								+ lifecycle + "] lifecycle");
						kernel.invoke(method).on(target).getResult();
					} catch (Throwable t) {
						logger.warn("Failed to callback [" + method
								+ "] on [" + lifecycle + "] lifecycle", t);
						
						if (lifecycle != FAILED) {
							callback(target, FAILED);
						}
					}
				}
			}
		}
		
		lifecycles.put(target, lifecycle);
	}

	public Kernel getKernel() {
		return kernel;
	}

	public void setKernel(Kernel kernel) {
		this.kernel = kernel;
	}

	public Repository getRepository() {
		return repository;
	}

	public void setRepository(Repository repository) {
		this.repository = repository;
	}

	public Bus getBus() {
		return bus;
	}

	public void setBus(Bus bus) {
		this.bus = bus;
	}
	
}
