/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.management;

import java.lang.annotation.Documented;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

/**
 * {@code On} annotation is used to mark a Jazzmaster component method as the 
 * lifecycle callback method.
 * Lifecycle callback method can be invoked by {@code Manager} when component 
 * lifecycle phase is changed. You can declare a lifecycle callback method on 
 * your component as the following code: 
 * <pre>
 * package jazzmaster.hello;
 * 
 * import java.lang.reflect.Method;
 * import java.util.ArrayList;
 * import java.util.List;
 * 
 * import org.eiichiro.jazzmaster.deployment.Stage;
 * import org.eiichiro.jazzmaster.deployment.Staged;
 * import org.eiichiro.jazzmaster.management.Lifecycle;
 * import org.eiichiro.jazzmaster.management.On;
 * import org.eiichiro.jazzmaster.management.Scope;
 * import org.eiichiro.jazzmaster.management.Scoped;
 * import org.eiichiro.jazzmaster.service.Implementation;
 * import org.eiichiro.jazzmaster.service.Named;
 * 
 * {@code @Named("Hello")}
 * {@code @Scoped(Scope.APPLICATION)}
 * {@code @Staged(Stage.TEST)}
 * {@code @Implementation}
 * public class HelloImpl implements Hello {
 * 
 *     {@code @On(Lifecycle.STARTED)}
 *     public void start() {
 *         System.out.println();
 *         System.out.println("HelloImpl started");
 *         System.out.println();
 *     }
 * </pre>
 * A Jazzmaster component has six lifecycle phases. 
 * However, the timing of component lifecycle changing is different for each 
 * component type. The timing of lifecycle changing for each component type can 
 * be reffered to each component documentation :-P
 *
 * @see Lifecycle
 * @see Manager
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
@Documented
@Retention(RetentionPolicy.RUNTIME)
@Target(ElementType.METHOD)
public @interface On {

	/**
	 * The array of {@code Lifecycle} to callback the annotated method.
	 * As the default, this element is all of the {@code Lifecycle} phases.
	 * 
	 * @return The array of {@code Lifecycle} to callback the annotated method.
	 */
	Lifecycle[] value();
	
}
