/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.repository;

import java.io.Serializable;
import java.lang.annotation.Annotation;
import java.util.HashSet;
import java.util.Set;

/**
 * {@code Key} is the key to store &amp; find service definition (
 * {@code org.eiichiro.jazzmaster.service.Service} and its extensions) in 
 * service repository.
 * At deployment time, {@code Repository} stores service definition with 
 * generated {@code Key}s from all of the service implementation type 
 * hierarchies and binding annotations. (A service definition can be identified 
 * with immediate {@code Key} and a service instance can be identified with the 
 * service definition)
 *
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class Key<T> implements Serializable {

	private static final long serialVersionUID = 5921128456807163993L;
	
	private final Class<T> type;
	
	private final Set<Annotation> bindings;
	
	/**
	 * Constructs a new {@code Key} instance with the specified {@code Class}.
	 * 
	 * @param type The {@code Class} of service.
	 */
	public Key(Class<T> type) {
		this(type, new HashSet<Annotation>(0));
	}
	
	/**
	 * Constructs a new {@code Key} instance with the specified {@code Class} 
	 * and the set of {@code Binding}.
	 * 
	 * @param type The {@code Class} of service.
	 * @param bindings The {@code Set} of service's {@code Binding}.
	 */
	public Key(Class<T> type, Set<Annotation> bindings) {
		this.type = type;
		this.bindings = bindings;
	}
	
	/**
	 * Returns {@code true} if this {@code Key} instance contains the specified 
	 * one.
	 * This method considers that this {@code Key} instance contains the 
	 * specified one, when this instance's {@code #type} equals the specified 
	 * {@code Key}'s {@code #type} and the specified {@code Key}'s binding 
	 * annotations are all contained in this instance's binding annotations.
	 * 
	 * @param key The {@code Key} to be tested.
	 * @return {@code true} if this {@code Key} instance contains the specified 
	 * one.
	 */
	public boolean contains(Key<?> key) {
		if (!this.type.equals(key.type)) {
			return false;
		}
		
		if (this.bindings.size() < key.bindings.size()) {
			return false;
		}
		
		for (Annotation binding : key.bindings) {
			if (!this.bindings.contains(binding)) {
				return false;
			}
		}
		
		return true;
	}
	
	/**
	 * @return the type
	 */
	public Class<?> getType() {
		return type;
	}
	
	public Set<Annotation> getBindings() {
		return bindings;
	}
	
	/** Returns this {@code Key} instance {@code String} representation. */
	@Override
	public String toString() {
		StringBuilder stringBuilder = new StringBuilder();
		stringBuilder.append("Type [" + type + "]");
		
		if (bindings.size() > 0) {
			stringBuilder.append("; Binding annotations [" + bindings + "]");
		}
		
		return stringBuilder.toString();
	}
	
}
