/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.service;

import java.lang.annotation.Documented;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

import org.eiichiro.jazzmaster.deployment.ClasspathScanner;
import org.eiichiro.jazzmaster.deployment.Deployer;
import org.eiichiro.jazzmaster.deployment.Scanner;
import org.eiichiro.jazzmaster.deployment.Stage;
import org.eiichiro.jazzmaster.deployment.Staged;
import org.eiichiro.jazzmaster.management.Scope;
import org.eiichiro.jazzmaster.management.Scoped;

/**
 * {@code Implementation} is a stereotype to mark a service class as service 
 * implementation. In Jazzmaster SOA, you can configure a service with two 
 * ways, that are the declarative one and the programmable one. 
 * {@code Implementation} annotation supports the declarative service 
 * configuration. Jazzmaster SOA built-in {@code Deployer} and {@code Scanner} 
 * consider {@code Implementation} annotated class as a service implementation 
 * class, generate service definition and deploy automatically. Declarative 
 * service configuration style using {@code Implementation} is: 
 * <pre>
 * package jazzmaster.hello;
 * 
 * import org.eiichiro.jazzmaster.deployment.Stage;
 * import org.eiichiro.jazzmaster.deployment.Staged;
 * import org.eiichiro.jazzmaster.kernel.Injected;
 * import org.eiichiro.jazzmaster.kernel.Kernel;
 * import org.eiichiro.jazzmaster.kernel.Property;
 * import org.eiichiro.jazzmaster.management.Scope;
 * import org.eiichiro.jazzmaster.management.Scoped;
 * import org.eiichiro.jazzmaster.service.Consumes;
 * import org.eiichiro.jazzmaster.service.Implementation;
 * import org.eiichiro.jazzmaster.service.Named;
 * 
 * {@code @Named("Hello")}
 * {@code @Scoped(Scope.APPLICATION)}
 * {@code @Staged(Stage.TEST)}
 * {@code @Implementation}
 * public class HelloImpl implements Hello {
 * 
 *     {@code @Injected} private Kernel kernel;
 * 	
 *     {@code @Injected} {@code @Consumes} private MyService myService;
 * 	
 *     {@code @Injected} {@code @Consumes} {@code @Named("YourService1")} private YourService yourService;
 * 	
 *     {@code @Injected} {@code @Consumes} {@code @Yours} private YourService yourService2;
 * 	
 *     public Message sayHello() {
 *         System.out.println(myService.toString());
 *         System.out.println(yourService.toString());
 *         System.out.println(yourService2.toString());
 *         return kernel.assemble(Message.class).with(new Property&lt;String&gt;("Hello, Jazzmaster!")).getInstance();
 *     }
 * 
 * }
 * </pre>
 * <pre>
 * package jazzmaster.hello;
 * 
 * import java.lang.annotation.Retention;
 * import java.lang.annotation.RetentionPolicy;
 * 
 * import org.eiichiro.jazzmaster.kernel.Binding;
 * 
 * {@code @Retention(RetentionPolicy.RUNTIME)}
 * {@code @Binding}
 * public {@code @interface} Yours {}
 * </pre>
 * <pre>
 * package jazzmaster.hello;
 * 
 * import org.eiichiro.jazzmaster.service.Implementation;
 * 
 * {@code @Yours}
 * {@code @Implementation}
 * public class YourServiceImpl2 implements YourService {}
 * </pre>
 * You can configure your service easily and quickly with dependency style. But, 
 * declarative service configuration using {@code Implementation} is <b>static</b>.
 * So, you cannot configure service for each service <b>instances</b> or at 
 * runtime. For instance-level service configuration or runtime configuration, 
 * you should choose programmable configuration style, that is declaring 
 * {@code Service} extension class.
 * 
 * @see Deployer
 * @see Scanner
 * @see ClasspathScanner
 * @see Scope
 * @see Scoped
 * @see Stage
 * @see Staged
 * @see Service
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
@Documented
@Retention(RetentionPolicy.RUNTIME)
@Target(ElementType.TYPE)
public @interface Implementation {}
