/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.service;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang.ClassUtils;
import org.eiichiro.jazzmaster.bus.Bus;
import org.eiichiro.jazzmaster.bus.Interceptor;
import org.eiichiro.jazzmaster.deployment.Deployer;
import org.eiichiro.jazzmaster.deployment.Stage;
import org.eiichiro.jazzmaster.kernel.Assembler;
import org.eiichiro.jazzmaster.kernel.Binding;
import org.eiichiro.jazzmaster.kernel.Injected;
import org.eiichiro.jazzmaster.kernel.Kernel;
import org.eiichiro.jazzmaster.kernel.Property;
import org.eiichiro.jazzmaster.management.Configuration;
import org.eiichiro.jazzmaster.management.Lifecycle;
import org.eiichiro.jazzmaster.management.Manager;
import org.eiichiro.jazzmaster.management.Scope;
import org.eiichiro.jazzmaster.repository.Repository;

/**
 * {@code Service} is the Jazzmaster SOA service definition representation.
 * In Jazzmaster SOA, you can configure a service with two ways, that are the 
 * declarative one and the programmable one. {@code Service} supports the 
 * programmable service configuration. You can configure your service with the 
 * programmable way by declaring this extension class and overriding appropriate 
 * method. Programmable service configuration style using {@code Service} is (as 
 * compared with {@code Implementation}): 
 * <pre>
 * package jazzmaster.hello;
 * 
 * import java.util.HashSet;
 * import java.util.Set;
 * 
 * import org.eiichiro.jazzmaster.kernel.Property;
 * import org.eiichiro.jazzmaster.management.Scope;
 * import org.eiichiro.jazzmaster.service.Reference;
 * import org.eiichiro.jazzmaster.service.Service;
 * 
 * public class HelloService extends Service {
 * 
 *     {@code @Override}
 *     public Class&lt;?&gt; getImplementation() {
 *         return HelloImpl2.class;
 *     }
 * 
 *     {@code @Override}
 *     public String getName() {
 *         return "Hello";
 *     }
 * 
 *     {@code @Override}
 *     public Scope getScope() {
 *         return Scope.APPLICATION;
 *     }
 * 
 *     {@code @SuppressWarnings("unchecked")}
 *     {@code @Override}
 *     public Set&lt;Property&lt;?&gt;&gt; getProperties() {
 *         Set&lt;Property&lt;?&gt;&gt; properties = new HashSet&lt;Property&lt;?&gt;&gt;();
 *         properties.add(new Property&lt;Reference&gt;(new Reference(MyService.class)));
 *         properties.add(new Property&lt;Reference&gt;(new Reference(named("YourService1"))));
 *         properties.add(new Property&lt;Reference&gt;(new Reference(YourService.class, Yours.class)));
 *         return properties;
 *     }
 * 
 * }
 * </pre>
 * You can configure your service for instance-level and at runtime by using 
 * this way. However, programmable service configuration with {@code Service} is 
 * so troublesome. If you want to configure your service easily and quickly, you 
 * should choose the declarative configuration style, that is declaring 
 * {@code Implementation} annotated class.
 *
 * @see Deployer
 * @see Bus
 * @see Repository
 * @see Implementation
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public abstract class Service {

	/** The {@code Kernel} for programmable service configuration. */
	@Injected protected Kernel kernel;
	
	/** The {@code Stage} for programmable service configuration. */
	@Injected protected Stage stage;
	
	/** The {@code Configuration} for programmable service configuration. */
	@Injected protected Configuration configuration;
	
	/** The {@code Manager} for programmable service configuration. */
	@Injected protected Manager manager;
	
	/** The {@code Deployer} for programmable service configuration. */
	@Injected protected Deployer deployer;
	
	/** The {@code Repository} for programmable service configuration. */
	@Injected protected Repository repository;
	
	/** The {@code Bus} for programmable service configuration. */
	@Injected protected Bus bus;
	
	/**
	 * Returns the service implementation class.
	 * You must implement this method for declaring service extension class with 
	 * the programmable service configuration.
	 * 
	 * @return The service implementation class.
	 */
	public abstract Class<?> getImplementation();
	
	/**
	 * Returns the service name.
	 * In default implementation, this method returns the value of 
	 * {@code getClass().getName()}. If you want to return your own service name 
	 * with programmable way, you should <b>override</b> this method.
	 * 
	 * @return The service name.
	 */
	@Overridable public String getName() {
		return getClass().getName();
	}
	
	/**
	 * Returns the {@code Scope} this service is available.
	 * In default implementation, this method returns {@code Scope#DEPENDENT}. 
	 * If you want to return another scope with programmable way, you should 
	 * <b>override</b> this method.
	 * 
	 * @return The scope this service is available.
	 */
	@Overridable public Scope getScope() {
		return Scope.DEPENDENT;
	}
	
	/**
	 * Returns the {@code List} view of interface this service to be exposed.
	 * In default implementation, this method returns 
	 * {@code getImplementation().getInterfaces()}. If you want to return your 
	 * own service interfaces with programmable way, you should <b>override</b> 
	 * this method.
	 * 
	 * @return The interfaces this service to be exposed.
	 */
	@Overridable public Set<Class<?>> getInterfaces() {
		Set<Class<?>> interfaces = new HashSet<Class<?>>();
		
		for (Object object : ClassUtils.getAllInterfaces(getImplementation())) {
			Class<?> clazz = (Class<?>) object;
			interfaces.add(clazz);
		}
		
		return interfaces;
	}
	
	/**
	 * Returns the {@code Set} view of {@code Property} to assemble this service 
	 * instance.
	 * In default implementation, this method returns empty {@code Set}. If you 
	 * want to return your own properties with programmable way, you should 
	 * <b>override</b> this method.
	 * 
	 * @return The properties to assemble this service instance.
	 */
	@Overridable public Set<Property<?>> getProperties() {
		return new HashSet<Property<?>>(0);
	}
	
	/**
	 * Returns the {@code Set} view of {@code Binding} annotation to qualify 
	 * this service instance for dependency injection component binding.
	 * In default implementation, this method returns empty {@code Set}. If you 
	 * want to return your own binding annotations by programmable way, you should 
	 * <b>override</b> this method.
	 * 
	 * @see Binding
	 * @return The binding annotations to qualify this service instance for 
	 * dependency injection component binding.
	 */
	@Overridable public Set<Annotation> getBindings() {
		return new HashSet<Annotation>(0);
	}
	
	/**
	 * Returns the built-in {@code Provider} implementation.
	 * This method returns the service provider implementation that assembles 
	 * service implementation class instance with {@code Kernel} and 
	 * {@code Assembler}. If you want to return your own {@code Provider} 
	 * implementation, you should <b>override</b> this method.
	 * 
	 * @see Kernel
	 * @see Assembler
	 * @return The built-in service provider implementation.
	 */
	@Overridable public <S> Provider<S> getProvider() {
		return new Provider<S>() {

			public S get() {
				@SuppressWarnings("unchecked")
				S instance = (S) kernel.assemble(getImplementation()).getInstance();
				return instance;
			}
			
		};
	}
	
	/**
	 * Returns the {@code List} view of service interceptor to intercept the 
	 * specified service method.
	 * In default implementation, this method returns empty {@code List}. If you 
	 * want to return your own interceptor with programmable way, you should 
	 * <b>override</b> this method.
	 * 
	 * @see Interceptor
	 * @param method The method to be intercept.
	 * @return The interceptors to intercept the specified service method.
	 */
	@Overridable public List<Class<?>> getInterceptors(Method method) {
		return new ArrayList<Class<?>>(0);
	}
	
	/**
	 * Returns deployment stages the service implementation instance is deployed.
	 * 
	 * @return Deployment stages the service implementation instance is deployed.
	 */
	@Overridable public Set<Stage> getStages() {
		Set<Stage> stages = new HashSet<Stage>(Stage.values().length);
		
		for (Stage stage : Stage.values()) {
			stages.add(stage);
		}
		
		return stages;
	}
	
	/**
	 * Returns callback methods corresponding to the specified lifecycle.
	 * 
	 * @param lifecycle The lifecycle that callback methods are invoked.
	 * @return The callback methods corresponding to the specified lifecycle.
	 */
	@Overridable public Set<Method> getCallbacks(Lifecycle lifecycle) {
		return new HashSet<Method>(0);
	}
	
	public void setKernel(Kernel kernel) {
		this.kernel = kernel;
	}

	public void setStage(Stage stage) {
		this.stage = stage;
	}

	public void setConfiguration(Configuration configuration) {
		this.configuration = configuration;
	}

	public void setManager(Manager manager) {
		this.manager = manager;
	}

	public void setDeployer(Deployer deployer) {
		this.deployer = deployer;
	}

	public void setRepository(Repository repository) {
		this.repository = repository;
	}

	public void setBus(Bus bus) {
		this.bus = bus;
	}
	
	/**
	 * Generates {@code Typed} binding annotation instance from the specified 
	 * service definition class.
	 * 
	 * @param type The service definition class to generate {@code Typed} 
	 * binding annotation.
	 * @return The {@code Typed} binding annotation instance generated from the 
	 * specified service definition class.
	 */
	public static Typed typed(Class<? extends Service> type) {
		return new TypedImpl(type);
	}
	
	/**
	 * Generates {@code Named} binding instance from the specified service name.
	 * 
	 * @param name The service name to generate {@code Named} binding annotation.
	 * @return The {@code Named} binding annotation instance generated from the 
	 * specified service name.
	 */
	public static Named named(String name) {
		return new NamedImpl(name);
	}
	
}
