/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jazzmaster.service;

import java.lang.annotation.Annotation;

/**
 * {@code TypedImpl}
 *
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
@SuppressWarnings("all")
class TypedImpl implements Typed {

	private final Class<? extends Service> value;
	
	/**
	 * Constructs a new {@code TypedImpl} instance with the specified service 
	 * definition type.
	 * 
	 * @param value The service definition type.
	 */
	public TypedImpl(Class<? extends Service> value) {
		this.value = value;
	}
	
	/**
	 * Returns the service definition type referred by {@code Typed} binding 
	 * annotation.
	 */
	public Class<? extends Service> value() {
		return value;
	}
	
	/** Returns {@code Typed.class} as the annotation type. */
	public Class<? extends Annotation> annotationType() {
		return Typed.class;
	}
	
	/**
	 * Returns {@code true} if the specified object implements {@code Typed} 
	 * interface and the specified {@code Typed#value()} is equivalent to this 
	 * {@code #value()}.
	 * 
	 * @param obj The target object to compare with.
	 * @return {@code true} if the specified object is equivalent to this one.
	 * @see Annotation#equals(Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj instanceof Typed) {
			Typed typed = Typed.class.cast(obj);
			
			if (value.equals(typed.value())) {
				return true;
			}
		}
		
		return false;
	}

	/**
	 * Returns the hash code of this instance.
	 * 
	 * @see Annotation#hashCode()
	 * @return The hash code of this instance.
	 */
	@Override
	public int hashCode() {
		return (127 * "value".hashCode()) ^ value.hashCode();
	}

	/**
	 * Returns the {@code String} representation of this instance.
	 * 
	 * @see Annotation#toString()
	 * @return The {@code String} representation of this instance.
	 */
	@Override
	public String toString() {
		return "@" + Typed.class.getName() + "(value=" + value.getName() + ")";
	}
	
}
